# This file is part of pymadcad,  distributed under license LGPL v3
'''
	This module provide functions to generate racks and gears with many different shapes.

	The mainstream gears are involute gears (using involute of circles as contact curves). This is due to the standard shape of racks. And involute gears are the best choice for most designs, which is why the current implementation focusses on them.

	However the tools here are modular, which means you can combine them with your own functions to create customized gears.

	Examples
	--------

	As a first use, you may want to generate a fully finished spur gear
	If you do not specify optional arguments, the function will provide good defaults for it.

		>>> # fully featured gear
		>>> gear(step=3, z=12, depth=4, bore_radius=1.5)

	You may want to generate a more bizarre gear, with a a different hub for instance. You will then need to assemble a gear from its 3 components:  exterior (tooths), structure (between hub and exterior), and hub (what holds the gear on an axis)

		>>> # this assemble a gear specifying each independant sub-parts
		>>> ext_radius = (3*12)/(2*pi) - 3
		>>> int_radius = 4
		>>> geargather(
		...		gearexterior(repeat_circular(gearprofile(3, 30), 30), depth=4),
		...		gearstructure('rounded', ext_radius, int_radius, 2, patterns=6),
		...		my_hub_mesh,
		...		)

	For reuse in your custom functions, the functions used to generate the gears are exposed:

		>>> # this is the raw profile of a tooth
		>>> gearprofile(step=3, z=12)
'''

from .mathutils import *
from .mesh import Web, Wire, Mesh, web, wire
from .blending import junction, blendpair
from .generation import extrusion, revolution, repeat, extrans
from .primitives import Circle, ArcCentered, Segment
from .triangulation import triangulation
from .selection import *
from .rendering import show
from .cut import bevel, chamfer
from .boolean import intersection
from . import settings

from math import *
from functools import reduce, partial
from operator import add
from copy import copy



def rackprofile(step, h=None, offset=0, alpha=radians(20), resolution=None) -> Wire:
	''' Generate a 1-period tooth profile for a rack

		Parameters:

			step:		period length over the primitive line
			h:		    tooth half height
			offset:     rack reference line offset with the primitive line (as a distance)
			              - the primitive line is the adherence line with gears
			              - the reference line is the line half the tooth is above and half below
			alpha:		angle of the tooth sides, a.k.a  pressure angle of the contact
	'''
	if h is None:
		h = default_height(step, alpha)
	e = offset  # change name for convenience
	x = 0.5 + 2*e/step*tan(alpha)  # fraction of the tooth above the primitive circle

	return Wire([
		vec3(step*x/2 - tan(alpha) * ( h-e),   h-e,  0),
		vec3(step*x/2 - tan(alpha) * (-h-e),  -h-e,  0),
		vec3(step*(2-x)/2 + tan(alpha) * (-h-e),  -h-e,  0),
		vec3(step*(2-x)/2 + tan(alpha) * ( h-e),   h-e,  0),
		vec3(step*(2+x)/2 - tan(alpha) * ( h-e),   h-e,  0),
		], groups=['rack'])


def gearprofile(step, z, h=None, offset=0, alpha=radians(20), resolution=None, **kwargs) -> Wire:
	''' Generate a 1-period tooth profile for a straight gear


		Parameters:

			step:		period length over the primitive circle
			z:			number of tooth on the gear this profile is meant for
			h:			tooth half height
			offset:		offset distance of the matching rack profile (see above)
			alpha:		pressure angle in the contact
	'''
	if h is None:
		h = default_height(step, alpha)
	p = step*z / (2*pi)	# primitive circle
	c = p * cos(alpha)	# tangent circle to gliding axis

	e = offset  # change name for convenience
	#e = offset + 0.05*h  # the real offset is 5% (the 1*m, 1.25*m) when the standard says it is 0
	x = 0.5 + 2*offset/step*tan(alpha)  # fraction of the tooth above the primitive circle

	o0 = angle(involute(c, 0, tan(alpha)))	# offset of contact curve
	oi = atan((h-e)/p * tan(alpha))		# offset of interference curve

	l0 = involuteat(c, p+h+e)	# interval size of contact curve

	# if the tooth height is unreachable, find the intersection between the two contact curves
	t0 = -step/(2*p)*x - o0
	t = t0+l0
	if involute(c, t0, t)[1] > 0:
		# Newton solver method
		for i in range(5):
			f = sin(t) + cos(t)*(t0-t)
			J = - sin(t)*(t0-t)
			t -= f/J
		l0 = t-t0

	# if there is an interference curve
	interference = c > p-h+e
	if interference:
		# Newton solver method
		# to compute the parameters (t1, t2) of the intersection between contact line and interference line
		t0, ti = o0, oi
		# initial state
		t1 = t0 - tan(alpha)	# put contact line on primitive
		t2 = ti + sqrt(c**2 - (p-h+e)**2) /p	# put interference point on base circle
		for i in range(8):
			ct1, ct2 = cos(t1), cos(t2)
			st1, st2 = sin(t1), sin(t2)
			# function value
			f = (	c*vec2(ct1,st1) + c*(t0-t1)*vec2(-st1,ct1)
				+	(h-e-p)*vec2(ct2,st2) -p*(ti-t2)*vec2(-st2,ct2)
				)
			# jacobian matrix (f partial derivatives)
			J = mat2(
				-c*(t0-t1)*vec2(ct1,st1),
				p*(ti-t2)*vec2(ct2,st2) + (h-e)*vec2(-st2,ct2),
				)
			# iteration
			t1, t2 = vec2(t1,t2) - inverse(J)*f
		li = t2 - ti	# interval size of interference curve
		s0 = t0 - t1	# generation start of contact curve
	else:
		s0 = involuteat(c, p-h+e)

	pts = []
	n = 2 + settings.curve_resolution(h, step/p, resolution)	# number of points to place

	# parameter for first side
	place = step/(2*p)*x	# place of intersection with the primitive circle
	t0 = place + o0	# start of contact curve
	ti = place + oi	# start of interference curve
	# contact line
	for i in range(n+1):
		t = interpol1(t0-l0, t0-s0, i/n)
		v = involute(c, t0, t)
		pts.append(vec3(v,0))
	# interference line
	if interference:
		for i in range(n+1):
			t = interpol1(ti+li, ti, i/n)
			v = involuteof(p, ti, -h+e, t)
			pts.append(vec3(v,0))

	# parameters for second side
	place = step/(2*p)*(2-x)
	t0 = place - o0
	ti = place - oi
	# interference line
	if interference:
		for i in range(n+1):
			t = interpol1(ti, ti-li, i/n)
			v = involuteof(p, ti, -h+e, t)
			pts.append(vec3(v,0))
	# contact line
	for i in range(n+1):
		t = interpol1(t0+s0, t0+l0, i/n)
		v = involute(c, t0, t)
		pts.append(vec3(v,0))

	pts.append(angleAxis(step/p, vec3(0,0,1)) * pts[0])

	return Wire(pts, groups=['gear'])

def gearcircles(step, z, h=None, offset=0, alpha=radians(30)):
	''' return the convenient circles radius for a gear with the given parameters
		  return is `(primitive, base, bottom, top)`
	'''
	if h is None:
		h = default_height(step, alpha)
	e = h*e
	p = step*z / (2*pi)	# primitive circle
	c = p * cos(alpha)	# tangent circle to gliding axis
	return p, c, p-h-e, p+h-e

def default_height(step, alpha):
	return 0.5 * 2.25 * step/pi * cos(alpha)/cos(radians(20))

def involute(c, t0, t):
	''' give a point of parameter `t` on involute from circle or radius `c`, starting from `t0` on the circle

		`t` and `t0` are angular positions
	'''
	x = vec2(cos(t), sin(t))
	y = vec2(-x[1], x[0])
	return c*(x + y*(t0-t))

def involuteat(c, r):
	''' give the parameter for the involute of circle radius `c` to reach radius `r` '''
	return sqrt((r/c)**2 - 1)

def involuteof(c, t0, d, t):
	''' give a point of parameter `t` on involute with offset, from circle or radius `c`, starting from `t0` on the circle

		`t` and `t0` are angular positions
	'''
	x = vec2(cos(t), sin(t))
	y = vec2(-x[1], x[0])
	return (c+d)*x + c*y*(t0-t)


def angle(p):
	return atan2(p[1], p[0])


def repeat_circular(profile, n: int) -> Wire:
	''' Repeat n times the given Wire by rotation around (O,Z) '''
	result = repeat(profile, n, rotatearound(
		anglebt(noproject(profile[0],Z), noproject(profile[-1],Z)), 
		(O,Z),
		))
	result.mergeclose()
	return result


def pattern_full(ext_radius, int_radius, depth, int_height=0, **kwargs) -> Mesh:
	"""
	Generate two full parts of the structure (the top and the bottom).

	Return a tuple (Web, Web, None) where the first `Web` is the top of the structure and
	the second `Web` is the bottom of the structure.

	Parameters:

		ext_radius (float): float (radius of the external border of the pattern
		int_radius (float): radius of the internal border of the pattern
		depth (float): face width
		int_height (float):

			if you want a pinion with a structure thinner than the value of `depth`,
			the total height will be `total_height = depth - 2 * int_height`
	"""
	half_depth = depth / 2
	assert half_depth > int_height, "`int_height` must be smaller than `depth / 2`"

	axis = (O, Z)
	circles_ref = web(Circle(axis, ext_radius)) + web(Circle(axis, int_radius)).flip()
	top_profile = circles_ref.transform(vec3(0, 0, half_depth - int_height))
	bottom_profile = circles_ref.transform(vec3(0, 0, -half_depth + int_height)).flip()
	mesh = triangulation(top_profile) + triangulation(bottom_profile)
	mesh.mergeclose()
	return mesh


def pattern_circle(
	ext_radius,
	int_radius,
	depth,
	int_height = 0,
	ratio = 1,
	patterns: int = 5,
	circles_radius = None,
	circles_place = None,
	**kwargs,
) -> Mesh:
	"""
	Generate two parts of the structure (the top and the bottom) with `patterns` distributed on the whole structure.

	Return a tuple (Web, Web, Mesh) where the first `Web` is the top of the structure,
	the second `Web` is the bottom of the structure and the last element `Mesh` is all side surfaces.

	Parameters:

		ext_radius (float): radius of the external border of the structure
		int_radius (float): radius of the internal border of the structure
		depth (float): face width
		int_height (float): if you want a pinion with a structure thinner than the value of `depth`,
							the total height will be `total_height = depth - 2 * int_height`
		ratio (float): number that allows to change proportionally the radius of circles
		patterns (int): number of circles of the structure
		circles_radius (float): radius of circles
		circles_place (float): radius where the origins of circles are placed

	Note:

		- for instance, with a ratio of 1.5, the radius of circles `circles_radius` will be divided by 1.5
		- if `circles_radius` is chosen, `ratio` won't impact the radius of circles `circles_radius`
	"""
	# Parameters
	half_depth = depth / 2
	if not (circles_radius) and not (circles_place):
		assert 0.5 < ratio, "`ratio` must be in the interval ]0.5; +inf["
	if not circles_place:
		circles_place = (ext_radius + int_radius) / 2
	if not circles_radius:
		circles_radius = (ext_radius - int_radius) / (4 * ratio)

	assert circles_radius / circles_place < sin(pi / patterns), "too many circles for the specified size. Change either 'ratio', 'circle_radius' or 'patterns'"

	Z = vec3(0,0,1)
	axis = (vec3(0,0,0), Z)

	# Borders
	ext_circle = web(Circle(axis, ext_radius))
	int_circle = web(Circle(axis, int_radius)).flip()

	# Pattern
	circle_ref = web(Circle((vec3(circles_place, 0, 0), Z), circles_radius))
	angle = 2 * pi / patterns
	pattern_profile = repeat(circle_ref.flip(), patterns, rotatearound(angle, axis))

	# Profiles and surfaces
	webs_ref = (pattern_profile, ext_circle, int_circle)
	top_webs = [wire.transform(vec3(0, 0, half_depth - int_height)) for wire in webs_ref]
	bottom_webs = [wire.transform(vec3(0, 0, -half_depth + int_height)).flip() for wire in webs_ref]

	top_profile = reduce(add, top_webs)
	bottom_profile = reduce(add, bottom_webs)
	surfaces = extrusion(vec3(0, 0, depth - 2 * int_height), bottom_webs[0].flip())

	mesh = triangulation(top_profile) + triangulation(bottom_profile) + surfaces
	mesh.mergeclose()
	return mesh


def create_pattern_rect(
	ext_radius,
	int_radius,
	depth,
	int_height = 0,
	ratio = 1,
	patterns: int = 5,
	ext_thickness = None,
	int_thickness = None,
	rounded: bool = False,
	**kwargs,
) -> Mesh:
	"""
	Function used for `pattern_rect` when `pattern` = "rect"
	and `pattern_rounded` when `pattern` = "rounded"

	Parameters:

		ext_radius (float): radius of the external border of the structure
		int_radius (float):  radius of the internal border of the structure
		depth (float): face width
		int_height (float): if you want a pinion with a structure thinner than the value of `depth`,
							the total height will be `total_height = depth - 2 * int_height`
		ratio (float): it is a number which is the proportional factor for an homothety of patterns
		patterns (int): number of patterns inside the structure
		ext_thickness (float): internal radius of the pattern
		int_thickness (float): external radius of the pattern
		rounded (bool): if it is True, the pattern will be rounded

	Note:

		- for instance, if `ratio` is 1.5, the area of the pattern will be divided by 1.5.
		- if `r_int` and `r_ext` are chosen, `ratio` won't impact these parameters
	"""
	# Parameters
	half_depth = depth / 2
	offset = ratio * 0.05 * (ext_radius - int_radius)
	r_ext = ext_radius - (ext_thickness or offset)
	r_int = int_radius + (int_thickness or offset)
	thickness = ratio * 0.15 * r_ext  # thickness of "arms"
	# angle between the bottom edge of an arm and the middle of it
	theta_1 = asin(thickness / (2 * r_int))
	# angle between the top edge of an arm and the middle of it
	theta_2 = asin(thickness / (2 * r_ext))
	angle_step = 2 * pi / patterns
	O = vec3(0)
	Z = vec3(0, 0, 1)
	axis = (O, Z)

	# Borders
	ext_circle = web(Circle(axis, ext_radius))
	int_circle = web(Circle(axis, int_radius)).flip()

	#function to convert cylindrical coordinates to cartesian coordinates
	cyl2cart = lambda r, theta: vec3(r * cos(theta), r * sin(theta), 0)
	# Vertex points of the pattern
	A1 = cyl2cart(r_int, theta_1)
	A2 = cyl2cart(r_ext, theta_2)
	B1 = cyl2cart(r_int, angle_step - theta_1)
	B2 = cyl2cart(r_ext, angle_step - theta_2)

	# Primitives
	pattern_ref = web([
					Segment(B2, B1),
					ArcCentered((O, -Z), B1, A1),
					Segment(A1, A2),
					ArcCentered((O, Z), A2, B2),
					])

	# Multiply the pattern of reference
	if rounded:  # rounded case
		bevel(	pattern_ref, 
				pattern_ref.frontiers(0,1,2), 
				('radius', min(0.3*(r_ext - r_int), r_int*(angle_step/2-theta_1))),
				)
		bevel(	pattern_ref, 
				pattern_ref.frontiers(2,3,0), 
				('radius', min(0.3*(r_ext - r_int), r_ext*(angle_step/2-theta_2))),
				)
		pattern_ref.mergeclose()
	pattern_profile = repeat(web(pattern_ref), patterns, rotatearound(angle_step, axis))

	# Profiles and surfaces
	webs_ref = (pattern_profile.flip(), ext_circle, int_circle)
	top_webs = [web.transform(vec3(0, 0, half_depth - int_height)) for web in webs_ref]
	bottom_webs = [web.transform(vec3(0, 0, -half_depth + int_height)).flip() for web in webs_ref]

	top_profile = reduce(add, top_webs)
	bottom_profile = reduce(add, bottom_webs)
	surfaces = extrusion(vec3(0, 0, depth - 2 * int_height), bottom_webs[0].flip())

	mesh = triangulation(top_profile) + triangulation(bottom_profile) + surfaces
	mesh.mergeclose()
	return mesh


def pattern_rect(
	ext_radius,
	int_radius,
	depth,
	int_height = 0,
	ratio = 1,
	patterns = 5,
	ext_thickness = None,
	int_thickness = None,
	**kwargs,
) -> Mesh:
	"""
	Generate two parts of the structure (the top and the bottom) with `patterns` distributed on the whole structure.
	All corners are straight. Check the function `pattern_rounded` to get rounded corners.

	Return a tuple (Web, Web, Mesh) where the first `Web` is the top of the structure,
	the second `Web` is the bottom of the structure and the last element `Mesh` is all side surfaces.

	Parameters:

		ext_radius: float (radius of the external border of the structure)
		int_radius: float (radius of the internal border of the structure)
		depth: float (face width)
		int_height: float (if you want a pinion with a structure thinner than the value of `depth`,
							the total height will be `total_height = depth - 2 * int_height`)
		ratio: float (it is a number which is the proportional factor for an homothety of patterns)
		patterns: int (number of patterns inside the structure)
		ext_thickness (float): internal radius of the pattern
		int_thickness (float): external radius of the pattern

	Note:

		- for instance, if `ratio` is 1.5, the area of the pattern will be divided by 1.5.
		- if `r_int` and `r_ext` are chosen, `ratio` won't impact these parameters
	"""
	return create_pattern_rect(ext_radius, int_radius, depth, int_height, ratio, patterns, ext_thickness, int_thickness, False)


def pattern_rounded(
	ext_radius,
	int_radius,
	depth,
	int_height = 0,
	ratio = 1,
	patterns: int = 5,
	ext_thickness = None,
	int_thickness = None,
	**kwargs,
) -> Mesh:
	"""
	Generate two parts of the structure (the top and the bottom) with `patterns` distributed on the whole structure.
	All corners are rounded. Check the function `pattern_rect` to get straight corners.

	Return a tuple (Web, Web, Mesh) where the first `Web` is the top of the structure,
	the second `Web` is the bottom of the structure and the last element `Mesh` is all side surfaces.

	Parameters:

		ext_radius: float (radius of the external border of the structure)
		int_radius: float (radius of the internal border of the structure)
		depth: float (face width)
		int_height: float (if you want a pinion with a structure thinner than the value of `depth`,
							the total height will be `total_height = depth - 2 * int_height`)
		ratio: float (it is a number which is the proportional factor for an homothety of patterns)
		patterns: int (number of patterns inside the structure)
		ext_thickness (float): internal radius of the pattern
		int_thickness (float): external radius of the pattern

	Note:

		- for instance, if `ratio` is 1.5, the area of the pattern will be divided by 1.5.
		- if `r_int` and `r_ext` are chosen, `ratio` won't impact these parameters
	"""
	return create_pattern_rect(ext_radius, int_radius, depth, int_height, ratio, patterns, ext_thickness, int_thickness, True)


def minmax_radius(points):
	min, max = inf, 0
	for p in points:
		r = length2(vec2(p))
		if r > max:		max = r
		elif r < min:	min = r
	return sqrt(min), sqrt(max)



def gearexterior(
	profile: Wire,
	depth,
	step = None,
	helix_angle = 0,
	chamfer = 0,
	**kwargs,
) -> Mesh:
	"""
	Generate the external part of the pinion

	Parameters:

		profile (Web):         profile of the pinion generated by `gearprofile`
		depth (float):         extrusion eight - width of the gear along its axis
		step (float):          step of chordal pitch, must be specified for non-null helix_angle, unused otherwise
		helix_angle (float):   helix angle for helical gears - only without bevel; `bevel` = False - it must be a radian angle
		chamfer (float):       chamfer angle - only for straight pinion
	"""
	# TODO: helical + chamfer

	half_depth = depth / 2
	footer, header = minmax_radius(profile.points)
	
	# Internal circles
	circle_ref = wire(Circle((O,Z), footer - 0.3*(header-footer)))
	top_circle = circle_ref.transform(vec3(0, 0, half_depth))
	bottom_circle = circle_ref.transform(vec3(0, 0, -half_depth))

	assert not (chamfer and helix_angle),  "chamfer is not (yet) supported with a non-null helix_angle"

	if chamfer:  # chamfer case
		# create a truncated profile (the top profile of the chamfer)
		start = mix(footer, header, 0.6)  # primitive radius
		k = tan(chamfer)
		truncated = deepcopy(profile) # profile which will be truncated
		for i, point in enumerate(profile.points):
			# truncate points that overlap the primitive
			r = length2(point)
			if r > start**2:  
				r = sqrt(r)
				profile.points[i] = point - vec3(0, 0, (r - start) * k)
				truncated.points[i] = point * start / r

		# Profiles
		top_profile = profile.transform(vec3(0, 0, half_depth))
		bottom_profile = top_profile.transform(scaledir(Z, -1))
		top_truncated = truncated.transform(vec3(0, 0, half_depth))
		bottom_truncated = truncated.transform(vec3(0, 0, -half_depth))

		# Surfaces
		bp = partial(junction, tangents="straight", resolution = ("div", 0))
		top_surface = triangulation(web([top_truncated, top_circle.flip()]))
		bottom_surface = triangulation(web([bottom_truncated.flip(), bottom_circle]))
		top_bevel = bp(top_truncated, top_profile.flip())
		bottom_bevel = bp(bottom_truncated.flip(), bottom_profile)
		gear_surface = bp(top_profile, bottom_profile.flip())

		surfaces = (top_surface, bottom_surface, top_bevel, bottom_bevel, gear_surface)
		mesh = reduce(add, surfaces)
		mesh.mergeclose()
	else:
		if helix_angle:  # helical teeth case
			# Edges of teeth
			R = header # the largest radius of profile
			# step to generate a helical transformation
			step = settings.curve_resolution(depth / cos(helix_angle), depth * tan(helix_angle) / R)
			angle = depth * tan(helix_angle) / R / (step + 1)
			h = depth / (step + 1)
			bottom_gear_edge = profile.transform(vec3(0, 0, -half_depth))
			transformations = (
				transform((vec3(0, 0, i * h), angleAxis(angle * i, Z)))
				for i in range(step + 2)
			)
			links = ((i, i + 1, 0) for i in range(step + 1))
			gear_surface = extrans(bottom_gear_edge, transformations, links)
			t = transform((vec3(0, 0, depth), angleAxis(depth * tan(helix_angle) / R, Z)))
			top_gear_edge = bottom_gear_edge.transform(t)
		else:  # straight teeth case
			# Edges of teeth
			bottom_gear_edge = profile.transform(vec3(0, 0, -half_depth))
			top_gear_edge = profile.transform(vec3(0, 0, half_depth))
			gear_surface = extrusion(vec3(0, 0, depth), bottom_gear_edge)

		# Surfaces
		top_surface = triangulation(web([top_gear_edge, top_circle.flip()]))
		bottom_surface = triangulation(web([bottom_gear_edge.flip(), bottom_circle]))
		mesh = gear_surface + top_surface + bottom_surface
		mesh.mergeclose()
	return mesh


def gearstructure(
	pattern,
	ext_radius,
	int_radius,
	depth,
	int_height = 0,
	ratio = 1,
	**kwargs,
) -> Mesh:
	"""
	Generate the internal part of the pinion

	Parameters:

		ext_radius (float):

				given by the attribut `_radius` of the result of `repeat_circular` -
				to avoid interference, it must be smaller than `_radius` (for instance 0.95 * `_radius`))

		int_radius (float):    it is the same radius of the largest radius of the hub part
		depth (float):         face width
		pattern:               any of 'full', 'circle', 'rect', 'rounded'
		int_height (float):

				if you want a pinion with a structure thinner than the value of `depth`,
				the total height will be `total_height = depth - 2 * int_height`
	"""
	# int_radius must not be 0
	int_radius = int_radius or 0.1 * ext_radius
	pattern_function = globals()['pattern_' + pattern]
	return pattern_function(ext_radius, int_radius, depth, int_height, ratio=ratio, **kwargs)


def gearhub(
	bore_radius,
	depth,
	int_height = 0,
	hub_height = None,
	hub_radius = None,
	**kwargs,
) -> Web:
	"""
	Generate a hub for a pinion part

	Parameters:

		bore_radius (float):     radius of the central bore
		depth (float):           face width; same parameter for `gearexterior` and `gearstructure`
		int_height (float):      only useful for no hub case, checkout the function `gearstructure` for more information
		hub_height (float):      height of the hub
		hub_radius (float):      external radius of the hub

	Note:

		- if `bore_radius` is null, the function will return a top circle and a bottom circle used for `geargather` function
		- if `hub_height` is null, the function will return a structure with a bore and without a hub
	"""
	if not (bore_radius):  # No hub case
		half_depth = depth / 2
		circle_ref = web(Circle((O,Z), depth * 0.1))
		circle_ref = triangulation(circle_ref)
		top_surface = circle_ref.transform(vec3(0, 0, half_depth - int_height))
		bottom_surface = circle_ref.transform(vec3(0, 0, -half_depth + int_height)).flip()
		return top_surface + bottom_surface

	if hub_height is None:
		hub_height = bore_radius
	if not (hub_radius):
		hub_radius = 2 * bore_radius
	height = hub_height + depth / 2 if hub_height else depth / 2 - int_height
	axis = (vec3(0, 0, height), vec3(0, 0, 1))
	bore_circle = Circle(axis, bore_radius)
	hub_circle = Circle(axis, hub_radius)

	# Top part
	top_surface = triangulation(web(bore_circle).flip() + web(hub_circle))

	# Lateral surfaces
	lateral_surfaces = extrusion(vec3(0, 0, -hub_height - depth), web(bore_circle))

	# Bottom part
	axis = (vec3(0, 0, -depth / 2), vec3(0, 0, 1))
	bottom_bore_circle = Circle(axis, bore_radius)
	bottom_hub_circle = Circle(axis, hub_radius)
	bottom_web = web(bottom_bore_circle) + web(bottom_hub_circle).flip()
	bottom_surface = triangulation(bottom_web)

	return top_surface + bottom_surface + lateral_surfaces


def geargather(exterior, structure, hub) -> Mesh:
	"""
	Gather all parts: exterior, structure, and hub

	You can obtain them via the provided functions, or generate them yourself.
	"""
	exterior.mergeclose()
	structure.mergeclose()
	hub.mergeclose()

	# Parameters
	box = structure.box()
	height_top = box.max.z
	height_bot = box.min.z
	ext_radius = box.max.x
	int_radius = 0

	box = hub.box()
	height_h_top = box.max.z
	height_h_bot = box.min.z
	ext_h_radius = box.max.x

	# Select borderlines
	# int = internal / ext = external
	# e = exterior / s = structure / h = hub
	# top = top / bot = bottom
	circle_int_e_top = select(exterior, vec3(0, 0, height_top))
	circle_int_e_bot = select(exterior, vec3(0, 0, height_bot))
	circle_ext_s_top = select(structure, vec3(ext_radius, 0, height_top))
	circle_ext_s_bot = select(structure, vec3(ext_radius, 0, height_bot))
	circle_int_s_top = select(structure, vec3(int_radius, 0, height_top))
	circle_int_s_bot = select(structure, vec3(int_radius, 0, height_bot))
	circle_ext_h_top = select(hub, vec3(ext_h_radius, 0, height_h_top))
	circle_ext_h_bot = select(hub, vec3(ext_h_radius, 0, height_h_bot))

	# Join all borderlines
	# j1 is the junction between `exterior` and `structure`
	# j2 is the junction between `structure` and `hub`
	j1_top = junction(circle_ext_s_top, circle_int_e_top, tangents="straight", resolution = ("div", 0))
	j1_bot = junction(circle_ext_s_bot, circle_int_e_bot, tangents="straight", resolution = ("div", 0))
	j2_top = junction(circle_ext_h_top, circle_int_s_top, tangents="straight", resolution = ("div", 0))
	j2_bot = junction(circle_ext_h_bot, circle_int_s_bot, tangents="straight", resolution = ("div", 0))
	j1 = j1_top + j1_bot
	j2 = j2_top + j2_bot

	mesh = exterior + j1 + structure + j2 + hub
	mesh.mergeclose()
	return mesh


def gear(
	step,
	z: int,
	depth,
	bore_radius = 0,
	int_height = 0,
	pattern = 'full',
	**kwargs,
) -> Mesh:
	"""
	Generate a pinion.

	Any extra argument will go to functions `gearprofile`, `gearstructure`, or `gearhub`

	Parameters:

		step (float):

			tooth step over the primitive curve, same as the matching rack step
			the primitive perimeter is `step * z` and radius is `step * z / 2*pi`

		z (int):         number of teeth
		depth (float):       extrusion eight - width of the gear along its axis
		bore_radius (float):   radius of the main bore
		int_height (float):    if you want a pinion with a structure thinner than the value of `depth`,
						       the total height will be `total_height = depth - 2 * int_height`
		pattern:
			determine the structure between exterior (tooth) and hub
			This argument specifies the use a a function named `'pattern_'+pattern` in this module.

	* Extra parameters for `gearprofile`

		offset (float):   offset of tooth (as a distance)
		alpha (float):    pressure angle in radian

	* Extra parameters for `gearexterior`

		helix_angle (float):   helix angle to get a helical pinion in radian
		chamfer (float):       chamfer angle - only for straight pinion

	* Extra parameters for `gearstructure`

		ratio (float):       influence the proportion of dimensions of the structure

		Note: `int_height` impacts the thickness of the structure unless specified

	* Extra parameters for `gearhub`

		hub_height (float):   height of the hub shoulder

	Note:

		- `int_height` impacts the height of the hub unless specified.
		- if `hub_height` is null, there will be no hub.
	"""
	profile = repeat_circular(gearprofile(step, z, **kwargs), z)

	# Parts
	exterior = gearexterior(profile, depth, step, **kwargs)
	hub = gearhub(bore_radius, depth, int_height, **kwargs)
	ext_int = minmax_radius(exterior.points)[0]
	structure = gearstructure(
					pattern,
					ext_int * 0.95,
					max(minmax_radius(hub.points)[1], 0.2*ext_int),
					depth,
					int_height,
					**kwargs)
	return geargather(exterior, structure, hub)


def frange(start:float, end:float, div:int=10):
	"""Generate `div` numbers from `start` to `end`"""
	k = (end - start) / (div - 1)
	return (start + i * k for i in range(div))


def get_pitch_cone_angle(z_pinion:int, z_wheel:int, shaft_angle:float=0.5 * pi) -> float:
	"""
	Return the pitch cone angle of the pinion called `gamma_p`.
	The pitch cone angle of the wheel is equal to `shaft_angle - gamma_p`

	Parameters:

		z_pinion (int): 		the number of teeth on the bevel pinion
		z_wheel (int): 			the number of teeth on the bevel wheel
		shaft_angle (float): 	the shaft angle
	"""
	return atan2(sin(shaft_angle), ((z_wheel / z_pinion) + cos(shaft_angle)))


def spherical_involute(cone_angle:float, t0:float, t:float) -> vec3:
	"""
	Return spherical involute function

	Parameters:

		t (float): 				the angular position
		t0 (float): 			the difference phase
		cone_angle (float): 	the cone angle

	Return:	
	
		a normalized `vec3`
	"""
	cos_g, sin_g = cos(cone_angle), sin(cone_angle)
	return vec3(
		sin_g * cos(t * sin_g) * cos(t + t0) + sin(t * sin_g) * sin(t + t0),
		sin_g * cos(t * sin_g) * sin(t + t0) - sin(t * sin_g) * cos(t + t0),
		cos_g * cos(t * sin_g),
	)


def spherical_involuteof(pitch_cone_angle:float, t0:float, alpha:float, t:float) -> vec3:
	"""
	Return the spherical interference function

	Parameters:

		t (float): 					the angular position
		t0 (float): 				the difference phase
		pitch_cone_angle (float):	the pitch cone angle
		alpha(float): 				the height angle offset of the rack

	Return:	
	
		a normalized `vec3`
	"""
	cos_p, sin_p = cos(pitch_cone_angle), sin(pitch_cone_angle)
	involute = lambda t, t0: spherical_involute(pitch_cone_angle, t0, t)
	vec = lambda t: vec3(-cos_p * cos(t), -cos_p * sin(t), sin_p)
	return cos(alpha) * involute(t, t0) + sin(alpha) * vec(t + t0)


def derived_spherical_involute(cone_angle:float, t0:float):
	"""
	Return the function of the derived spherical involute function.

	Parameters:

		cone_angle (float): 	the cone angle
		t0 (float): 			the phase difference
	"""
	cos_g, sin_g = cos(cone_angle), sin(cone_angle)
	return lambda t: vec3(
		cos_g ** 2 * sin(t * sin_g) * cos(t + t0),
		cos_g ** 2 * sin(t * sin_g) * sin(t + t0),
		-cos_g * sin_g * sin(t * sin_g),
	)


def jacobian_spherical_involute(base_cona_angle:float, pitch_cone_angle:float, t01:float, t02:float, alpha:float) -> callable:
	"""
	Return the function of the jacobian used for the newton method in `spherical_gearprofile`

	Parameters:

		base_cona_angle (float): 	the base cone angle
		pitch_cone_angle (float): 	the pitch cone angle
		t01 (float): 				the phase of the spherical involute function
		t02 (float): 				the phase of the spherical interference function
		alpha (float): 				the height angle offset of the rack
	"""
	dsi = derived_spherical_involute # for convenience
	derived_involute = dsi(base_cona_angle, t01)
	cos_p = cos(pitch_cone_angle)
	vec = lambda t: vec3(cos_p * sin(t), -cos_p * cos(t), 0)
	derived_interference = lambda t: dsi(pitch_cone_angle, t02)(t) * cos(alpha) + sin(alpha) * vec(t + t02)
	return lambda t1, t2: mat3(derived_involute(t1), -derived_interference(t2), vec3(0, 0, 1))


def spherical_rack_tools(z:float, pressure_angle:float=pi / 9, ka:float=1, kd:float=1.25):
	"""
	Return a list of all information useful to generate a spherical rack.
	Five elements :
	
		1) the minimum abscissa for the function (fifth element)
		2) the maximum abscissa for the function (fifth element)
		3) the phase of a tooth
		4) the phase of space
		5) the function to generate a tooth

	Parameters :

		z (float):
			
			number of tooth of the rack equal to `z_pinion / sin(pitch_cone_angle)`
			or `z_wheel / sin(shaft_angle - pitch_cone_angle)`
		
		pressure_angle (float): the pressure angle of the gear
		ka (float): 			the addendum coefficient
		kd (float): 			the dedendum coefficient
	"""
	k = 1 / z
	gamma_p = 0.5 * pi
	gamma_b = asin(cos(pressure_angle))
	cos_b, sin_b = cos(gamma_b), sin(gamma_b)
	gamma_f = gamma_p + 2 * ka * k
	gamma_r = gamma_p - 2 * kd * k

	phi_p = acos(tan(gamma_b) / tan(gamma_p))
	theta_p = atan2(sin_b * tan(phi_p), 1) / sin_b - phi_p
	phase_diff = k * pi + 2 * theta_p

	t_min = acos(cos(gamma_r) / cos_b) / sin_b
	t_max = acos(cos(gamma_f) / cos_b) / sin_b

	involute = lambda t, t0: spherical_involute(gamma_b, t0, t)
	v = vec3(1, 1, 0)
	phase_empty = 2 * pi * k - anglebt(involute(t_min, 0) * v, involute(-t_min, phase_diff) * v)

	return [t_min, t_max, phase_diff, phase_empty, involute]


def spherical_rackprofile(z:float, pressure_angle:float=pi / 9, ka:float=1, kd:float=1.25):
	"""
	Return a `Wire` which is a tooth of the rack.

	Parameters:

		z (float):
		
			number of tooth of the rack equal to `z_pinion / sin(pitch_cone_angle)`
			or `z_wheel / sin(shaft_angle - pitch_cone_angle)`
		
		pressure_angle (float): 	the pressure angle of the gear
		ka (float): 				the addendum coefficient
		kd (float): 				the dedendum coefficient
	"""
	t_min, t_max, phase1, phase2, involute = spherical_rack_tools(z, pressure_angle, ka, kd)
	side1 = [involute(t, 0) for t in frange(t_min, t_max)]
	side2 = [involute(-t, phase1) for t in frange(t_min, t_max)]
	segment = Segment(involute(t_min, -phase2), side1[0]).mesh()
	segment2 = Segment(side1[-1], side2[-1]).mesh()
	wire = segment + Wire(side1) + segment2 + Wire(side2).flip()
	return wire


def spherical_gearprofile(z:int, pitch_cone_angle:float, pressure_angle:float=pi / 9, ka:float=1, kd:float=1.25) -> Wire:
	"""
	Generate and return a `Wire` of a 1-period tooth spherical profile for a bevel gear

	Parameters:

		z (int):						number of tooth on the gear this profile is meant for
		pitch_cone_angle (float): 		the pitch cone angle
		pressure_angle (float):			pressure angle of the tooth
		ka (float):						addendum coefficient
		kd (float): 					dedendum coefficient
	"""
	# Initialization of parameters
	gamma_p = pitch_cone_angle # for convenience
	gamma_b = asin(cos(pressure_angle) * sin(gamma_p))
	cos_b, sin_b = cos(gamma_b), sin(gamma_b)
	tooth_size = pi / z
	involute = lambda t, t0 : spherical_involute(gamma_b, t0, t)
	epsilon_p = acos(cos(gamma_p) / cos_b) / sin_b
	theta_p = anglebt(involute(0, 0) * vec3(1, 1, 0), involute(epsilon_p, 0) * vec3(1, 1, 0))
	phase_diff = tooth_size + 2 * theta_p
	phase_empty = phase_interference = 2 * pi / z - phase_diff
	# The following number `k` is useful to simplify some calculations
	# It's broadly speaking `1/z_rack` and `z_rack` is not an integer !
	k = sin(gamma_p) / z

	# Spherical involute part
	gamma_f = gamma_p + 2 * ka * k # addendum cone angle
	gamma_r = gamma_p - 2 * kd * k # dedendum cone angle
	t_min = 0
	t_max = acos(cos(gamma_f) / cos_b) / sin_b
	if gamma_r > gamma_b:
		v = vec3(1, 1, 0)
		t_min = acos(cos(gamma_r) / cos_b) / sin_b
		phase_empty = 2 * pi / z - anglebt(
			involute(t_min, 0) * v, involute(-t_min, phase_diff) * v
		)

	# Calculation of offsets due to geometry of spherical rack
	_, t_rack_max, phase1, _, rinvolute = spherical_rack_tools(1 / k, pressure_angle, ka, kd)
	interference = lambda t, t0 : spherical_involuteof(gamma_p, t0, alpha, t)
	alpha = 2 * ka * k
	n1, n2 = rinvolute(t_rack_max, 0) * vec3(1, 1, 0), rinvolute(-t_rack_max, phase1) * vec3(1, 1, 0)
	beta = 0.5 * anglebt(n1, n2) * length(n1) / sin_b

	# Newton method to calculate the intersection between
	# the spherical involute and the spherical interference.
	# Objective function
	involuteat = lambda t2, t0 : spherical_involuteof(gamma_p, t0, alpha, t2)
	f = lambda t1, t2: involute(t1, 0) - involuteat(t2, -0.5 * phase_interference + beta)
	# Jacobian matrix
	J = jacobian_spherical_involute(gamma_b, gamma_p, 0, -0.5 * phase_interference + beta, alpha)

	# Compute the intersection values
	t1, t2, t3 = 0.5 * t_max, -0.5 * t_max, 0
	for i in range(8):
		t1, t2, t3 = vec3(t1, t2, t3) - inverse(J(t1, t2)) * f(t1, t2)

	# Build sides of a tooth
	interference1 = [interference(t, -0.5 * phase_interference + beta) for t in frange(0, t2)]
	interference2 = [interference(-t, phase_diff + 0.5 * phase_interference - beta) for t in frange(0, t2)]
	side1 = interference1[:-1] + [involute(t, 0) for t in frange(t1, t_max)]
	side2 = interference2[:-1] + [involute(-t, phase_diff) for t in frange(t1, t_max)]

	# Extreme points of sides to compute angle between them
	a = interference(0, -0.5 * phase_interference + beta)
	b = interference(0, phase_diff + 0.5 * phase_interference - beta)
	final_phase_empty = 2 * pi / z - anglebt(a * vec3(1, 1, 0), b * vec3(1, 1, 0))
	top = Segment(involute(t_max, 0), involute(-t_max, phase_diff)).mesh()
	bottom = Segment(angleAxis(-final_phase_empty, vec3(0, 0, 1)) * a, a).mesh()

	return bottom + Wire(side1) + top + Wire(side2).flip()


def cone_projection(profile: Wire, pitch_cone_angle:float) -> Wire:
	"""
	Return a Wire of the spherical profile projected on a cone

	Parameters:
		profile (Wire) : 			the spherical profile
		pitch_cone_angle (float) : 	the pitch cone angle
	"""
	ref = lambda t: vec3(sin(pitch_cone_angle) * cos(t), sin(pitch_cone_angle) * sin(t), cos(pitch_cone_angle))
	new_points = [1 / dot(ref(atan2(point.y, point.x)), point) * point for point in profile.points]
	return Wire(new_points, indices=profile.indices)


def bevelgear(step:float, z:int, pitch_cone_angle:float, pressure_angle:float=pi/9, ka:float=1, kd:float=1.25, bore_radius:float=None, bore_height:float=None):
	"""
	Generate a bevel gear.

	Parameters:

		step (float):

			tooth step over the primitive curve, same as the matching rack step
			the primitive perimeter is `step * z` and radius is `step * z / (2 * pi)`

		z (int):		 			number of teeth
		pitch_cone_angle (float):	   		pitch cone angle
		pressure_angle (flaot):		the pressure angle of the tooth
		bore_radius (float):   		radius of the main bore
		bore_height (float):   		height of the main bore
	"""

	# Initialization of parameters
	gamma_p = pitch_cone_angle # for convenience
	gamma_b = asin(cos(pressure_angle) * sin(gamma_p))
	sin_b = cos(pressure_angle) * sin(gamma_p)
	rp = z*step / (2*pi)
	rho1 = rp / sin(gamma_p)
	rho0 = 2 * rho1 / 3
	k = sin(gamma_p) / z
	gamma_r = gamma_p - 2 * kd * k
	phi_p = acos(tan(gamma_b) / tan(gamma_p))
	theta_p = atan2(sin_b * tan(phi_p), 1) / sin_b - phi_p
	phase_diff = pi / z + 2 * theta_p

	# Generate spherical profiles
	spherical_profile = spherical_gearprofile(z, gamma_p, pressure_angle, ka, kd) # one tooth
	outside_profile = spherical_profile.transform(rho1 * 1.1)
	inside_profile = spherical_profile.transform(rho0 * 0.9)

	# Generate teeth border
	teeth_border = blendpair(outside_profile, inside_profile.flip(), tangents="straight")

	# Common values
	v = vec3(1, 1, 0)
	angle1tooth = anglebt(spherical_profile[0] * v, spherical_profile[-1] * v)
	gamma_l = gamma_p + 2 * (ka + 0.25) * k # offset : 0.25 for boolean operation
	A = vec3(rho1 * sin(gamma_r), 0, rho1 * cos(gamma_r))
	B = vec3(rho1 * sin(gamma_l), 0, rho1 * cos(gamma_l))
	C = vec3(rho0 * sin(gamma_r), 0, rho0 * cos(gamma_r))
	D = vec3(rho0 * sin(gamma_l), 0, rho0 * cos(gamma_l))
	v1 = A * v
	v2 = spherical_profile[0] * v
	phase_tooth_body = anglebt(v1, v2)
	phase_tooth_body = -phase_tooth_body if cross(v1, v2).z > 0 else phase_tooth_body


	# Generate points for a section
	if bore_radius is None:
		bore_radius = 0.5 * rho0 * sin(gamma_r)
	if bore_height is None:
		bore_height = 0.4 * rho1 * cos(gamma_r)
	if bore_radius:
		if bore_height:
			E = vec3(1.5 * bore_radius, 0, rho1 * cos(gamma_r))
			F = vec3(bore_radius, 0, rho0 * cos(gamma_r))
			G = vec3(1.5 * bore_radius, 0, rho1 * cos(gamma_r) + bore_height) # top of the bore
			H = vec3(bore_radius, 0, rho1 * cos(gamma_r) + bore_height) # top of the bore
			wire = Wire([A, B, D, C, F, H, G, E, A]).segmented()
			chamfer(wire, [4, 5, 6], ("distance", bore_radius * 0.05))
			bevel(wire, [7], ("distance", bore_height * 0.1))
		else:
			E = vec3(bore_radius, 0, rho1 * cos(gamma_r))
			F = vec3(bore_radius, 0, rho0 * cos(gamma_r))
			wire = Wire([A, B, D, C, F, E, A]).segmented()
			chamfer(wire, [4, 5], ("distance", bore_radius * 0.05))
	else:
		E = vec3(0, 0, rho1 * cos(gamma_r))
		F = vec3(0, 0, rho0 * cos(gamma_r))
		wire = Wire([A, B, D, C, F, E, A]).segmented()

	axis = (O,Z)
	body = revolution(angle1tooth, axis, wire)
	one_tooth = intersection(body, teeth_border.transform(angleAxis(phase_tooth_body, axis[1])))
	all_teeth = repeat(one_tooth, z, rotatearound(angle1tooth, axis))
	all_teeth.finish()
	return all_teeth



''' pymadcad submodule for mesh post processing '''

def stretch(mesh, curve=0.1, locked=()):
	''' stretching smoothing of the surface until reaching a certain curve (rad/m)
		locked can be a set of points that won't be moved
		
		We advise having a well subdivided surface (use `subdivide` before if necessary)
	'''
	conn = connef(mesh.faces)
	edges = set(edgekey(*e)  for e in conn)
	pts = mesh.points
	def current_curve(a,b):
		c = arrangeface(mesh.faces[conn[(a,b)]], a)[2]
		d = arrangeface(mesh.faces[conn[(b,a)]], b)[2]
		z = normalise(b-a)
		cy = noproject(c-a, z)
		dy = noproject(d-a, z)
		cyl = length(cy)
		dyl = length(dy)
		return cross(dy, dy) / (cyl*dyl) / (cyl+dyl)
	
	curve = {e: current_curve(*e)	for e in edges}	# curve around each edge
	indev

def refine(mesh, ratio:float) -> Mesh:
	''' increase the resolution by adding interpolated points to the surface '''
	indev
	
def flipdiags(mesh) -> Mesh:
	''' flip quad diagonals to improve the surface smoothness '''
	indev
	
def subdivide(mesh, cuts=1) -> Mesh:
	''' subdivide each triangle '''
	indev
	
def edgedivide(mesh, edges=None, cuts=1) -> Mesh:
	''' subdivide the given edges in the mesh, or mesh outlines if nothing is provided '''
	if edges is None:	edges = mesh.outlines_oriented()
	indev





from .rendering import Group
from .kinematic import Solid
from .mesh import Container
from .mathutils import vec3
from .primitives import isaxis

def annotate(source, scene, recur=-1):
	if not isinstance(scene, Scene):
		raise TypeError('scene must be a Scene')
	
	def recur(source, level, recur):
		for sub,src in sourceitems(source):
			disp = level.displays.get(sub)
			if not disp:	continue
			
			if isinstance(disp, Group):
				annotate(src, disp.displays, recur-1, scene)
			
			if isinstance(src, vec3):
				level.displays[str(sub)+'#annotation'] = text.TextDisplay(
					scene, obj, str(sub), 
					size=settings.display['view_font_size'], 
					color=settings.display['annotation_color'], 
					align=(-1,1),
					)
			elif isaxis(src):
				pass # TODO
			elif isinstance(src, Container):
				pass # TODO: arrow annotation
			elif isinstance(src, Kinematic):
				makescheme(src.joints)
	
	recur(source, scene, recur)
	

def sourceitems(obj):
	if isinstance(obj, list):	return enumerate(obj)
	elif isinstance(obj, dict):	return obj.items()
	else:	return ()




# This file is part of pymadcad,  distributed under license LGPL v3

'''  This modules provide tools for accessing data using its space location (not for final user).
	
	The complexity and therefore the cost of those operations are most of the time close to the hashmap complexity O(1). It means data is found in time independently of the actual size of the mesh or whatever storage it is.
'''

from .mathutils import *
from . import core
from . import mesh
from functools import reduce
from math import floor, ceil, sqrt, inf


class PositionMap:
	''' Holds objects assoiated with their location
		every object can be bound to multiple locations, and each location can hold multiple objects
		cellsize defines the box size for location hashing (the smaller it is, the bigger the memory footprint will be for non-point primitives)
		
		Attributes defined here:
			:cellsize:    the boxing parameter (DON'T CHANGE IT IF NON-EMPTY)
			:dict:        the hashmap from box to objects lists
	'''
	__slots__ = 'cellsize', 'dict', 'options'
	def __init__(self, cellsize, iterable=None):
		self.options = {}
		self.cellsize = cellsize
		self.dict = {}
		if iterable:	self.update(iterable)

	def keysfor(self, space):
		''' rasterize the primitive, yielding the successive position keys 
			currently allowed primitives are 
				points:	vec3
				segments:  (vec3,vec3)
				triangles: (vec3,vec3,vec3)
		'''
		cell = self.cellsize
		# point
		if isinstance(space, vec3):
			yield tuple(i64vec3(glm.floor(space/cell)))
		
		# segment
		elif isinstance(space, tuple) and len(space) == 2:
			# permutation of coordinates to get the direction the closer to Z
			n = glm.abs(space[1]-space[0])
			if max(n) < NUMPREC*max((max(p) for p in space)):	return
			if   n[1] >= n[0] and n[1] >= n[2]:	order,reorder = (2,0,1),(1,2,0)
			elif n[0] >= n[1] and n[0] >= n[2]:	order,reorder = (1,2,0),(2,0,1)
			else:								order,reorder = (0,1,2),(0,1,2)
			space = [ vec3(p[order[0]], p[order[1]], p[order[2]])	for p in space]
			
			# prepare variables
			cell2 = cell/2
			v = space[1]-space[0]
			dy = v[1]/v[2]
			dx = v[0]/v[2]
			o = space[0]
			fy = lambda z:	o[1] + dy*(z-o[2])
			fx = lambda z:	o[0] + dx*(z-o[2])
			
			# z selection
			if v[2] > 0:	zmin,zmax = space[0][2], space[1][2]
			else:			zmin,zmax = space[1][2], space[0][2]
			zmin -= zmin%cell
			for i in range(max(1,ceil((zmax-zmin)/cell))):
				z = zmin + cell*i + cell2
				
				# y selection
				if dy > 0:	ymin,ymax = fy(z-cell2), fy(z+cell2)
				else:		ymin,ymax = fy(z+cell2), fy(z-cell2)
				ymin -= ymin%cell
				for j in range(max(1,ceil((ymax-ymin)/cell))):
					y = ymin + j*cell + cell2
					
					# x selection
					if dx > 0:	xmin,xmax = fx(z-cell2), fx(z+cell2)
					else:		xmin,xmax = fx(z+cell2), fx(z-cell2)
					xmin -= xmin%cell
					for k in range(max(1,ceil((xmax-xmin)/cell))):
						x = xmin + k*cell + cell2
						
						p = (x,y,z)
						yield tuple([floor(p[i]/cell)		for i in reorder])
		
		# triangle
		elif isinstance(space, tuple) and len(space) == 3:
			# permutation of coordinates to get the normal the closer to Z
			n = glm.abs(cross(space[1]-space[0], space[2]-space[0]))
			if max(n) < NUMPREC*max((max(p) for p in space)):	return
			if   n[1] >= n[0] and n[1] >= n[2]:	order,reorder = (2,0,1),(1,2,0)
			elif n[0] >= n[1] and n[0] >= n[2]:	order,reorder = (1,2,0),(2,0,1)
			else:								order,reorder = (0,1,2),(0,1,2)
			space = [ vec3(p[order[0]], p[order[1]], p[order[2]])	for p in space]
			
			# prepare variables
			v = [space[i-1]-space[i]	for i in range(3)]
			n = cross(v[0],v[1])
			dx = -n[0]/n[2]
			dy = -n[1]/n[2]
			o = space[0]
			fz = lambda x,y:	o[2] + dx*(x-o[0]) + dy*(y-o[1])
			cell2 = cell/2
			pmin = reduce(glm.min, space)
			pmax = reduce(glm.max, space)
			xmin,xmax = pmin[0],pmax[0]
			pmin -= pmin%cell
			pmax += cell - pmax%cell
			
			# x selection
			xmin -= xmin%cell
			for i in range(max(1,ceil((xmax-xmin)/cell))):
				x = xmin + cell*i + cell2
			
				# y selection
				cand = []
				for i in range(3):
					# NOTE: cet interval ajoute parfois des cases inutiles apres les sommets
					if (space[i-1][0]-x+cell2)*(space[i][0]-x-cell2) <= 0 or (space[i-1][0]-x-cell2)*(space[i][0]-x+cell2) <= 0:
						d = v[i][1]/(v[i][0] if v[i][0] else inf)
						cand.append( space[i][1] + d * (x-cell2-space[i][0]) )
						cand.append( space[i][1] + d * (x+cell2-space[i][0]) )
				ymin,ymax = max(pmin[1],min(cand)), min(pmax[1],max(cand))
				ymin -= ymin%cell
				for j in range(max(1,ceil((ymax-ymin)/cell))):
					y = ymin + cell*j + cell2
				
					# z selection
					cand = []
					cand.append( fz(x-cell2, y-cell2) )
					cand.append( fz(x+cell2, y-cell2) )
					cand.append( fz(x-cell2, y+cell2) )
					cand.append( fz(x+cell2, y+cell2) )
					zmin,zmax = max(pmin[2],min(cand)), min(pmax[2],max(cand))
					zmin -= zmin%cell
					for k in range(max(1,ceil((zmax-zmin)/cell))):
						z = zmin + cell*k + cell2
						
						# remove box from corners that goes out of the area
						p = (x,y,z)
						if pmin[0]<p[0] and pmin[1]<p[1] and pmin[2]<p[2] and p[0]<pmax[0] and p[1]<pmax[1] and p[2]<pmax[2]:
							yield tuple([floor(p[i]/cell)	for i in reorder])
		
		else:
			raise TypeError("PositionMap only supports keys of type:  points, segments, triangles")
	
	def keysfor(self, space):
		''' rasterize the primitive, yielding the successive position keys 
			currently allowed primitives are 
				:points: 	vec3
				:segments:   (vec3,vec3)
				:triangles:  (vec3,vec3,vec3)
		'''
		# point
		if isinstance(space, vec3):
			return tuple(i64vec3(glm.floor(space/cell))),
		# segment
		elif isinstance(space, tuple) and len(space) == 2:
			return core.rasterize_segment(space, self.cellsize)
		# triangle
		elif isinstance(space, tuple) and len(space) == 3:
			return core.rasterize_triangle(space, self.cellsize)
		else:
			raise TypeError("PositionMap only supports keys of type:  points, segments, triangles")
	
	def update(self, other):
		''' add the elements from an other PositionMap or from an iteravble '''
		if isinstance(other, PositionMap):
			assert self.cellsize == other.cellsize
			for k,v in other.dict.items():
				if k in self.dict:	self.dict[k].extend(v)
				else:				self.dict[k] = v
		elif hasattr(other, '__iter__'):
			for space,obj in other:
				self.add(space,obj)
		else:
			raise TypeError("update requires a PositionMap or an iterable of couples (space, obj)")
	
	def add(self, space, obj):
		''' add an object associated with a primitive '''
		for k in self.keysfor(space):
			if k not in self.dict:	self.dict[k] = [obj]
			else:					self.dict[k].append(obj)
	
	def get(self, space):
		''' get the objects associated with the given primitive '''
		for k in self.keysfor(space):
			if k in self.dict:
				yield from self.dict[k]
				
	def __contains__(self, space):
		return next(self.get(space), None) is not None
	
	_display = (
		[	
			vec3(0,0,0),
			vec3(1,0,0),vec3(0,1,0),vec3(0,0,1),
			vec3(0,1,1),vec3(1,0,1),vec3(1,1,0),
			vec3(1,1,1)],
		[	
			(0,1),(0,2),(0,3),
			(1,6),(2,6),
			(1,5),(3,5),
			(2,4),(3,4),
			(4,7),(5,7),(6,7),
			],
		)
	def display(self, scene):
		web = mesh.Web()
		if 'color' in self.options:		web.options['color'] = self.options['color']
		base = vec3(self.cellsize)
		for k in self.dict:
			l = len(web.points)
			web += mesh.Web([base*(p+k)  for p in self._display[0]], self._display[1], groups=[k])
		return web.display(scene)

def meshcellsize(mesh):
	''' returns a good cell size to index primitives of a mesh with a PositionMap 
		
		See implementation.
	'''
	# the number of key for triangles is propotionate to the surface
	# points are placed on this surface, so sqrt(len(pts)) is a good hint for the point density on the surface
	return length(mesh.box().width) / sqrt(len(mesh.points))


class PointSet:
	''' Holds a list of points and hash them.
		the points are holds using indices, that allows to get the point buffer at any time, or to retreive only a point index
		cellsize defines the box size in which two points are considered to be the same
		
		methods are inspired from the builtin type set
		
		Attributes defined here:
			:points:     the point buffer (READ-ONLY PURPOSE)
			:cellsize:   the boxing parameter (DON'T CHANGE IT IF NON-EMPTY). mendatory and is the distance at which you want to distinguish points
			:dict:       the hashmap from box to point indices
			
		Build parameters:
			:iterable:    use it to build the set by inserting elements
			:manage:      pass a list for inplace use it, only indexing will be built
	'''
	__slots__ = 'points', 'cellsize', 'dict'
	def __init__(self, cellsize, iterable=None, manage=None):
		self.cellsize = cellsize
		self.dict = {}
		if manage is not None:
			self.points = manage
			for i in reversed(range(len(self.points))):
				self.dict[self.keyfor(self.points[i])] = i
		else:
			self.points = typedlist(dtype=vec3)
			if iterable:	self.update(iterable)
	
	def keyfor(self, pt):
		''' hash key for a point '''
		return tuple(i64vec3(glm.floor(pt/self.cellsize)))
		
	def keysfor(self, pt):
		''' iterable of positions at whic an equivalent point can be '''
		vox = pt/self.cellsize
		k = i64vec3(glm.floor(vox-0.5+NUMPREC)), i64vec3(glm.floor(vox+0.5-NUMPREC))
		return (
			(k[0][0], k[0][1], k[0][2]),
			(k[1][0], k[0][1], k[0][2]),
			(k[0][0], k[1][1], k[0][2]),
			(k[1][0], k[1][1], k[0][2]),
			(k[0][0], k[0][1], k[1][2]),
			(k[1][0], k[0][1], k[1][2]),
			(k[0][0], k[1][1], k[1][2]),
			(k[1][0], k[1][1], k[1][2]),
			)
	
	def update(self, iterable):
		''' add the points from an iterable '''
		for pt in iterable:	self.add(pt)
	def difference_update(self, iterable):
		''' remove the points from an iteravble '''
		for pt in iterable:	self.discard(pt)
		
	def add(self, pt):
		''' add a point '''
		for key in self.keysfor(pt):
			if key in self.dict:
				return self.dict[key]
		self.dict[self.keyfor(pt)] = l = len(self.points)
		self.points.append(pt)
		return l
	def remove(self, pt):
		''' remove a point '''
		for key in self.keysfor(pt):
			if key in self.dict:
				del self.dict[key]
				return
		else:					raise IndexError("position doesn't exist in set")
	def discard(self, pt):
		''' remove the point at given location if any '''
		for key in self.keysfor(pt):
			if key in self.dict:
				del self.dict[key]
	
	def __contains__(self, pt):
		for key in self.keysfor(pt):
			if key in self.dict:	return True
		return False
	def __getitem__(self, pt):
		for key in self.keysfor(pt):
			if key in self.dict:	return self.dict[key]
		raise IndexError("position doesn't exist in set")
		
	__iadd__ = update
	__isub__ = difference_update
	def __add__(self, iterable):
		s = PositionSet()
		s.union(self.points)
		s.union(iterable)
		return s
	def __sub__(self, iterable):
		s = PositionSet()
		s.union(self.points)
		s.difference(iterable)
		return s




# This file is part of pymadcad,  distributed under license LGPL v3
'''
	This module provides functions to compute convex hulls 
	(See https://en.wikipedia.org/wiki/Convex_hull) 
	and other hull-related operations for `Mesh` and `Web`
	
	Those can be very helpful to complete sketches or parts by adding the missing surface portions. Also very helpful to sort a set of directions.
'''

import scipy.spatial
from .mathutils import *
from .mesh import Mesh, Web, Wire, edgekey, facekeyo, arrangeface, connef, connpe, numpy_to_typedlist, typedlist_to_numpy
from .asso import Asso

from copy import copy
from collections import Counter



def simple_convexhull(points: '[vec3]') -> '[uvec3]':
	''' Just like `convexhull()` but minimalist.
		It does not take care of groups crossing. It doesn't return a new Mesh but a buffer of triangles indices 
	'''
	return numpy_to_typedlist(scipy.spatial.ConvexHull(typedlist_to_numpy(points, 'f8'), qhull_options='QJ Pp').simplices, uvec3)
	

def convexhull(source: '[vec3]') -> Mesh:
	''' compute the convex hull of the input container 
		
		Parameters:
			source (typedlist/Web/Mesh):	the input container, if it is a Web or Mesh, their groups are kept in the output data

		Example:
			
			>>> m = convexhull(mesh([
			... 	uvsphere(O, 1, alignment=X),
			... 	uvsphere(4*X, 2, alignment=X),
			... 	]))
	'''
	if isinstance(source, (Mesh,Web,Wire)):
		source = copy(source)
		source.strippoints()
		points = source.points
	else:
		points = typedlist(source, dtype=vec3)
		source = None
	
	return restore_groups(source, simple_convexhull(points)).orient()
	

def convexoutline(source: '[vec3]', normal: vec3=None, flatten: bool=False) -> Web:
	''' based on `convexhull()` but will extract the loop formed by the edges in the biggest planar projection of the convex hull 
	
		Parameters:
			source (typedlist/Web/Mesh):     the input container, if it is a Web or Mesh, their groups are kept in the output data
			normal:                          the projection normal to retreive the outline using `horizon()`, if None is given it will default to the direction in which the outlien surface is the biggest
			flatten:                         whether to project the outline points in its mean plane
	
		Example:
			
			>>> convexoutline(web([
			... 	Circle(Axis(O,Z), 1),
			... 	Circle(Axis(4*X,Z), 2),
			... 	]))
	'''
	hull = convexhull(source)
	direction = normal or widest_surface_direction(hull)
	outline = horizon(hull, direction)
	if flatten:
		outline.strippoints()
		center = outline.barycenter()
		outline.points = typedlist(center + noproject(p-center, direction)   for p in outline.points)
	return outline
	

def restore_groups(source, indices) -> Mesh:	
	'''
		Create a mesh contaning the given simplices, associated with groups created from group crossing from the source mesh.
		The simplices must refer to points in `source`.
		The created groups are tuples of indices of the groups touched by each simplex.
	
		Parameters:
			source (Mesh/Web):    the mesh containing reference triangles and groups
			indices:              a buffer of simplices indices (depending on the dimensionnality of `source`) we want to find groups for
	'''
	groups = source.groups
	tracks = typedlist.full(len(groups), len(indices), dtype='I')
	
	if isinstance(source, Mesh):
		former = { facekeyo(*face): track   
						for face, track in zip(source.faces, source.tracks)}
	elif isinstance(source, Web):
		former = { edge: track
						for edge, track in zip(source.edges, source.tracks)}
	else:
		raise TypeError('expected a Mesh or Web')
	
	# create an associative dictionnary where associated tracks only appears one per point
	combiner = Asso() 
	for simplex, track in former.items():
		for p in simplex:
			if track not in combiner[p]:
				combiner.add(p, track)
	# search for original group or new group at each simplex
	groupindex = {}
	for i, simplex in enumerate(indices):
		belong = Counter()
		for p in simplex:
			belong.update(combiner[p])
		# get a shortlist of the groups reaching this simplex through its points
		couple = sorted(belong, key=lambda i: belong[i], reverse=True)
		shortlist = []
		amount = 0
		while amount < len(simplex):
			n = couple[len(shortlist)]
			shortlist.append(n)
			amount += belong[n]
		# one group only 
		if len(shortlist) == 1:
			tracks[i] = couple[0]
		else:
			combine = tuple(sorted(shortlist))
			if combine not in groupindex:
				groupindex[combine] = len(groups)
				groups.append(combine)
			tracks[i] = groupindex[combine]
	
	outdim = len(simplex)
	if outdim == 3:
		return Mesh(source.points, indices, tracks, groups)
	elif outdim == 2:
		return Web(source.points, indices, tracks, groups)
	else:
		raise TypeError('outer simplex dimension {} is not supported'.format(outdim))

		
def horizon(mesh, direction: vec3) -> Web:
	''' 
		Return a Web of the ORIENTED edges of the given mesh that lay between triangles that are oriented either sides of `direction`
	'''
	horizon = Web(points=mesh.points, groups=mesh.groups)
	signs = {}	# dictionnary for crossing face directions around edges
	groups = {} # track of the positive face connected for each edge
	for face, track in zip(mesh.faces, mesh.tracks):
		# pick the sign of the projected surface
		s = dot(mesh.facenormal(face), direction)
		if abs(s) <= NUMPREC:	s = 0
		elif s > 0:				s = 1
		else:					s = -1
		
		for i in range(3):
			edge = (face[i], face[i-2])
			key = edgekey(*edge)
			# pick the group of the positive face neighboring the edge
			if s > 0 or (s >= 0 and key not in groups):
				groups[key] = track
			# the edge belong to the horizon if the neigboring faces are each appart the projection plane
			if edge in signs and signs[edge] * s <= 0 and max(signs[edge], s) == 1:
				horizon.edges.append(edge if s > 0 else flipped(edge))
				horizon.tracks.append(groups[key])
				del signs[edge]
				del groups[key]
			# if matching edge not found yet, index the current face
			else:
				signs[flipped(edge)] = s
	return horizon

	
def flipped(simplex):
	if len(simplex) == 2:
		return (simplex[1], simplex[0])
	elif len(simplex) == 3:
		return (simplex[2], simplex[1], simplex[2])
	else:
		raise TypeError('expected and edge or a triangle')
	
	
def widest_surface_direction(mesh) -> vec3:
	# find a direction not orthogonal to any of the mesh faces
	normals = mesh.facenormals()
	directionmap = simple_convexhull(normals)
	score = 0
	proj = None
	for face in directionmap:
		surf = length2(cross(normals[face[1]]-normals[face[0]], normals[face[2]]-normals[face[0]]))
		if surf >= score:
			score, proj = surf, normals[face[0]] + normals[face[1]] + normals[face[2]]
	
	# project half of the mesh surface
	direction = vec3(0)
	for face in mesh.faces:
		surf = cross(mesh.points[face[1]]-mesh.points[face[0]], mesh.points[face[2]]-mesh.points[face[0]])
		if dot(surf, proj) > 0:
			direction += surf
	
	# select the average surface direction
	return normalize(direction)
	
	
"""
def hull(bounds: Web) -> Web:
	conn = connpe(bounds)
	bridges = line_bridges(bounds, conn)
	# select a start point the most outward
	center = bounds.barycenter()
	start = min( (i  for e in bounds.edges for i in e), 
				key=lambda i: distance2(bounds.points[i], center))
	indev
	
def hull(bounds: Mesh) -> Mesh:
	indev
	
def minkowski(a: Mesh, b: Mesh) -> Mesh:
	''' minkowski sum of the input meshes '''
	indev
	
def brush(brush: Mesh, path, orient=None) -> Mesh:
	''' almost the same as a minkowski sum, but allows for orientation changes of the brush mesh, for each vertex of the path mesh 
	
		Parameters:
			brush (Mesh/Web):	the input mesh to move along `path` and being reoriented by `orient`
			path (Mesh/Web):	the input set to move the `brush` along, can be a Web or a Mesh, independently of the dimension of `brush`
			orient (typedlist):	
				list of vectors matching the point buffer of `path`, and giving the orientation that `brush` is taking in each of the `path` points
				If set to `None`, it defaults to the path normals
	'''
	indev
"""




# This file is part of pymadcad,  distributed under license LGPL v3

from math import log, exp, floor
from .mathutils import (vec3, fvec3, fvec4, fmat4, 
						mix, normalize, length, distance, dot, noproject, dirbase, transform,
						Box, isnan, isinf,
						)
from .rendering import Display, overrides, writeproperty
from .common import ressourcedir
from . import settings
from . import primitives
from PIL import Image
import numpy.core as np
import moderngl as mgl

from PyQt5.QtCore import Qt, QEvent


def shader_wire(scene):
	return scene.ctx.program(
				vertex_shader=open(ressourcedir+'/shaders/wire.vert').read(),
				fragment_shader=open(ressourcedir+'/shaders/wire.frag').read(),
				)
	
def shader_uniformcolor(scene):
	return scene.ctx.program(
				vertex_shader=open(ressourcedir+'/shaders/uniformcolor.vert').read(),
				fragment_shader=open(ressourcedir+'/shaders/uniformcolor.frag').read(),
				)


class PointDisplay(Display):
	def __init__(self, scene, position, size=10, color=None):
		self.position = fvec3(position)
		self.size = size
		self.selected = False
		self.color = fvec3(color or settings.display['line_color'])
		
		def load(scene):
			img = Image.open(ressourcedir+'/textures/point.png')
			texture = scene.ctx.texture(img.size, 1, img.convert('L').tobytes())
			#self.texture = scene.ressource('pointtex', load)
			shader = scene.ctx.program(
						vertex_shader=open(ressourcedir+'/shaders/pointhalo.vert').read(),
						fragment_shader=open(ressourcedir+'/shaders/pointhalo.frag').read(),
						)
			shader['halotex'].value = 0
			ident_shader = scene.ctx.program(
						vertex_shader=open(ressourcedir+'/shaders/pointhalo-ident.vert').read(),
						fragment_shader=open(ressourcedir+'/shaders/ident.frag').read(),
						)
			#self.shader = scene.ressource('pointhalo', load)
			vb = scene.ctx.buffer(np.array([(0,0), (0,1), (1,1), (0,0), (1,1), (1,0)], 'f4'))
			va = scene.ctx.vertex_array(shader, [(vb, '2f', 'v_uv')])
			va_ident = scene.ctx.vertex_array(ident_shader, [(vb, '2f', 'v_uv')])
			return texture, shader, va, ident_shader, va_ident
		
		(	self.texture, 
			self.shader, 
			self.va, 
			self.ident_shader, 
			self.va_ident	) = scene.ressource('pointhalo', load)
			
	@property
	def box(self):
		return Box(center=self.position, width=fvec3(0))
	
	def render(self, view):
		self.shader['color'].write(fvec3(settings.display['select_color_line']) if self.selected else self.color)
		self.shader['position'].write(fvec3(self.world * fvec4(self.position,1)))
		self.shader['view'].write(view.uniforms['view'])
		self.shader['proj'].write(view.uniforms['proj'])
		self.shader['ratio'] = (
				self.size / view.width(),
				self.size / view.height(),
				)
		self.texture.use(0)
		self.va.render(mgl.TRIANGLES)

	def identify(self, view):
		self.ident_shader['ident'] = view.identstep(1)
		self.ident_shader['position'].write(fvec3(self.world * fvec4(self.position,1)))
		self.ident_shader['view'].write(view.uniforms['view'])
		self.ident_shader['proj'].write(view.uniforms['proj'])
		self.ident_shader['ratio'] = (
				1.5 * self.size / view.width(),
				1.5 * self.size / view.height(),
				)
		self.va_ident.render(mgl.TRIANGLES)
	
	def stack(self, scene):
		return ( ((), 'ident', 2, self.identify),
				 ((), 'screen', 2, self.render))


class AxisDisplay(Display):
	pattern = [0, 0.25, 0.45, 0.55, 0.75, 1]
	repetitions = 3
	def __init__(self, scene, axis, interval=None, color=None, pose=fmat4(1)):
		self.origin = fvec3(axis[0])
		self.direction = fvec3(axis[1])
		self.interval = interval
		self.color = fvec3(color or settings.display['line_color'])
		self.selected = False
		self.box = Box(center=self.origin, width=fvec3(0))
		
		self.shader, self.va, self.ident_shader, self.va_ident = scene.ressource('axis', self.load)
	
	def load(self, scene):
		shader = scene.ctx.program(
					vertex_shader=open(ressourcedir+'/shaders/axis.vert').read(),
					fragment_shader=open(ressourcedir+'/shaders/axis.frag').read(),
					)
		ident_shader = scene.ctx.program(
					vertex_shader=open(ressourcedir+'/shaders/axis-ident.vert').read(),
					fragment_shader=open(ressourcedir+'/shaders/ident.frag').read(),
					)
		pts = []
		for i in range(-1, self.repetitions+1):
			for pt in self.pattern:
				if i == -1:					alpha = pt
				elif i == self.repetitions:	alpha = 1-pt
				else:						alpha = 1
				pts.append(((pt+i)/self.repetitions, alpha))
		vb = scene.ctx.buffer(np.array(pts, 'f4'))
		va = scene.ctx.vertex_array(shader, [(vb, 'f f', 'v_absciss', 'v_alpha')])
		va_ident = scene.ctx.vertex_array(ident_shader, [(vb, 'f 12x', 'v_absciss')])
		return shader, va, ident_shader, va_ident
	
	def _disp_interval(self, view):
		if self.interval:	return self.interval
		else:
			size = -view.uniforms['view'][3][2]/6
			return (-0.5*size, size)
	
	def render(self, view):
		self.shader['projview'].write(view.uniforms['projview'])
		self.shader['world'].write(self.world)
		self.shader['origin'].write(self.origin)
		self.shader['direction'].write(self.direction)
		self.shader['interval'] = self._disp_interval(view)
		self.shader['color'].write(fvec3(settings.display['select_color_line']) if self.selected else self.color)
		self.va.render(mgl.LINES)
	
	def identify(self, view):
		self.ident_shader['projview'].write(view.uniforms['projview'])
		self.ident_shader['world'].write(self.world)
		self.ident_shader['origin'].write(self.origin)
		self.ident_shader['direction'].write(self.direction)
		self.ident_shader['interval'] = self._disp_interval(view)
		self.ident_shader['ident'] = view.identstep(1)
		self.va_ident.render(mgl.LINES)
	
	def stack(self, scene):
		return ( ((), 'ident', 2, self.identify),
				 ((), 'screen', 2, self.render))

def npboundingbox(points):
	''' boundingbox for numpy arrays of points on the 3 first components '''
	return Box(
				fvec3([float(np.min(points[:,i])) for i in range(3)]),
				fvec3([float(np.max(points[:,i])) for i in range(3)]),
				)

class AnnotationDisplay(Display):
	def __init__(self, scene, points, color):
		self.color = fvec3(color or settings.display['annotation_color'])
		self.selected = False
		self.box = npboundingbox(points)
		# load shader
		def load(scene):
			return scene.ctx.program(
						vertex_shader=open(ressourcedir+'/shaders/annotation.vert').read(),
						fragment_shader=open(ressourcedir+'/shaders/annotation.frag').read(),
						)
		self.shader = scene.ressource('shader_annotation', load)
		# allocate buffers
		self.vb_pts = scene.ctx.buffer(points)
		self.va = scene.ctx.vertex_array(self.shader, [(self.vb_pts, '3f f', 'v_position', 'v_alpha')])
		self.va_ident = scene.ctx.vertex_array(scene.ressource('shader_ident'), [(self.vb_pts, '3f 4x', 'v_position')])
		
	def __del__(self):
		self.va.release()
		self.va_ident.release()
		self.vb_pts.release()
	
	@staticmethod
	def buff_ptsalpha(points, alpha):
		return np.hstack((
				np.array([tuple(p) for p in points], 'f4'), 
				np.array(alpha, 'f4', ndmin=2).transpose(),
				))

	def render(self, view):
		self.shader['proj'].write(view.uniforms['proj'])
		self.shader['view'].write(view.uniforms['view'] * self.world)
		self.shader['color'].write(self.color if not self.selected else fvec3(settings.display['select_color_line']))
		self.va.render(mgl.LINES)
	
	def identify(self, view):
		shader = view.scene.ressource('shader_ident')
		shader['proj'].write(view.uniforms['proj'])
		shader['view'].write(view.uniforms['view'] * self.world)
		shader['ident'] = view.identstep(1)
		self.va_ident.render(mgl.LINES)
		
	def stack(self, scene):
		return ( ((), 'ident', 2, self.identify),
				 ((), 'screen', 2, self.render)) 


class BoxDisplay(AnnotationDisplay):
	def __init__(self, scene, box, color=None):
		# place points
		x,y,z = box.width
		c = 0.1*min(x,y,z)	# corner
		o = 0.4 # wire alpha
		pts = np.array([
			# corner
			(0,0,0,1),(c,0,0,1),
			(0,0,0,1),(0,c,0,1),
			(0,0,0,1),(0,0,2*c,1),
			
			# box wire
			(0,0,0,o),(x,0,0,o),
			(0,0,0,o),(0,y,0,o),
			(0,0,0,o),(0,0,z,o),
			
			(x,0,0,o),(x,y,0,o),
			(x,0,0,o),(x,0,z,o),
			(0,y,0,o),(x,y,0,o),
			(0,y,0,o),(0,y,z,o),
			(0,0,z,o),(x,0,z,o),
			(0,0,z,o),(0,y,z,o),
			
			(x,y,0,o),(x,y,z,o),
			(x,0,z,o),(x,y,z,o),
			(0,y,z,o),(x,y,z,o),
			], dtype='f4')
		pts += (*box.min, 0)
		self.textplace = box.min

		super().__init__(scene, pts, color)
	

class SolidDisplay(Display):
	''' Display render Meshes '''
	def __init__(self, scene, positions, normals, faces, lines, idents, color=None):
		self.box = npboundingbox(positions)
		self.options = scene.options
		color = fvec3(color or settings.display['solid_color'])
		setting = settings.display['line_color']
		line = length(setting) * normalize(mix(color, setting, 0.2))	if max(color) else setting
		reflect = normalize(color) * settings.display['solid_reflectivity']
		self.vertices = Vertices(scene.ctx, positions, idents)
		self.disp_faces = FacesDisplay(scene, self.vertices, normals, faces, color=color, reflect=reflect, layer=0)
		self.disp_ghost = GhostDisplay(scene, self.vertices, normals, faces, color=line, layer=0)
		self.disp_groups = LinesDisplay(scene, self.vertices, lines, color=line, alpha=1, layer=-2e-6)
		self.disp_points = PointsDisplay(scene, self.vertices, range(len(positions)), layer=-3e-6)
		wire = []
		for f in faces:
			wire.append((f[0], f[1]))
			wire.append((f[1], f[2]))
			wire.append((f[2], f[0]))
		self.disp_wire = LinesDisplay(scene, self.vertices, wire, color=line, alpha=0.3, layer=-1e-6)
		
	def stack(self, scene):
		yield ((), 'screen', -1, self.vertices.prerender)
		if self.options['display_faces']:	
			yield ((), 'screen', 0, self.disp_faces.render)
			yield ((), 'ident', 0, self.disp_faces.identify)
		else:								
			yield ((), 'screen', 1, self.disp_ghost.render)
			yield ((), 'ident', 0, self.disp_ghost.identify)
		if self.options['display_groups']:	yield ((), 'screen', 1, self.disp_groups.render)
		if self.options['display_points']:	yield ((), 'screen', 2, self.disp_points.render)
		if self.options['display_wire']:	yield ((), 'screen', 2, self.disp_wire.render)
	
	@property
	def world(self):	return self.vertices.world
	@world.setter
	def world(self, value):	self.vertices.world = value

	#def control(self, view, key, sub, evt):
		#if evt.type() == QEvent.MouseButtonRelease and evt.button() == Qt.LeftButton:
			#sub = sub[0]
			#flags, idents = self.vertices.flags, self.vertices.idents
			#for i in range(len(idents)):
				#flags[i] ^= idents[i] == sub
			#self.vertices.flags_updated = True
			#view.update()
			#evt.accept()
	

class WebDisplay(Display):
	''' Display to render Webs '''
	def __init__(self, scene, positions, lines, points, idents, color=None):
		self.box = npboundingbox(positions)
		self.options = scene.options
		color = color or settings.display['line_color']
		self.vertices = Vertices(scene.ctx, positions, idents)
		self.disp_edges = LinesDisplay(scene, self.vertices, lines, color=color, alpha=1, layer=-2e-6)
		self.disp_groups = PointsDisplay(scene, self.vertices, points, layer=-3e-6)
		self.disp_points = PointsDisplay(scene, self.vertices, range(len(positions)), layer=-1e-6)
		

	def stack(self, scene):
		yield ((), 'screen', -1, self.vertices.prerender)
		if self.options['display_groups']:		yield ((), 'screen', 2, self.disp_groups.render)
		if self.options['display_points']:		yield ((), 'screen', 2, self.disp_points.render)
		yield ((), 'screen', 1, self.disp_edges.render)
		yield ((), 'ident', 1, self.disp_edges.identify)
	
	@property
	def world(self):	return self.vertices.world
	@world.setter
	def world(self, value):	self.vertices.world = value

	#def control(self, view, key, sub, evt):
		#if evt.type() == QEvent.MouseButtonRelease and evt.button() == Qt.LeftButton:
			#sub = sub[0]
			#flags, idents = self.vertices.flags, self.vertices.idents
			#for i in range(len(idents)):
				#flags[i] ^= idents[i] == sub
			#self.vertices.flags_updated = True
			#view.update()
			#evt.accept()


class Vertices(object):
	''' convenient class to share vertices between SolidDisplay, WebDisplay, PointsDisplay '''
	def __init__(self, ctx, positions, idents):
		self.idents = idents
		self.nident = int(max(idents))+1
		self.flags = np.zeros(len(positions), dtype='u1')
		self.flags_updated = False
		assert len(idents) == len(positions)
		self.vb_positions = ctx.buffer(np.array(positions, dtype='f4', copy=False))
		self.vb_idents = ctx.buffer(np.array(idents, dtype='u2', copy=False))
		self.vb_flags = self.vb_flags = ctx.buffer(self.flags, dynamic=True)
		self.world = fmat4(1)
		
	def __del__(self):
		self.vb_positions.release()
		self.vb_idents.release()
		self.vb_flags.release()
		
	def prerender(self, view):
		if self.flags_updated:
			self.vb_flags.write(self.flags)
			self.flags_updated = False
	
	def selectsub(self, sub):
		for i,id in enumerate(self.idents):
			self.flags[i] ^= id == sub
		self.flags_updated = True
			


class FacesDisplay:
	def __init__(self, scene, vertices, normals, faces, color, reflect, layer=0):
		self.color = color
		self.layer = layer
		self.reflect = reflect
		self.vertices = vertices
	
		# load the skybox texture
		def load(scene):
			img = Image.open(ressourcedir+'/textures/'+settings.display['solid_reflect'])
			return scene.ctx.texture(img.size, 3, img.tobytes())
		self.reflectmap = scene.ressource('skybox', load)
		
		# load the shader
		def load(scene):
			shader = scene.ctx.program(
						vertex_shader=open(ressourcedir+'/shaders/solid.vert').read(),
						fragment_shader=open(ressourcedir+'/shaders/solid.frag').read(),
						)
			# setup some uniforms
			shader['reflectmap'] = 0
			return shader
		self.shader = scene.ressource('shader_solid', load)
		self.ident_shader = scene.ressource('shader_subident')
		# allocate buffers
		if faces is not None and len(faces) and vertices.vb_positions:
			self.vb_faces = scene.ctx.buffer(np.array(faces, 'u4', copy=False))
			self.vb_normals = scene.ctx.buffer(np.array(normals, 'f4', copy=False))
			self.va = scene.ctx.vertex_array(
					self.shader, 
					[	(vertices.vb_positions, '3f', 'v_position'), 
						(self.vb_normals, '3f', 'v_normal'),
						(vertices.vb_flags, 'u1', 'v_flags')],
					self.vb_faces,
					)
			
			self.va_ident = scene.ctx.vertex_array(
					self.ident_shader, 
					[	(vertices.vb_positions, '3f', 'v_position'),
						(vertices.vb_idents, 'u2', 'item_ident')], 
					self.vb_faces,
					)
		else:
			self.va = None
			
	def __del__(self):
		if self.va:
			self.va.release()
			self.va_ident.release()
			self.vb_faces.release()
			self.vb_normals.release()
	
	def render(self, view):
		if self.va:
			# setup uniforms
			self.shader['select_color'].write(settings.display['select_color_face'])
			self.shader['min_color'].write(self.color * settings.display['solid_color_side'])
			self.shader['max_color'].write(self.color * settings.display['solid_color_front'])
			self.shader['refl_color'].write(self.reflect)
			self.shader['layer'] = self.layer
			self.shader['world'].write(self.vertices.world)
			self.shader['view'].write(view.uniforms['view'])
			self.shader['proj'].write(view.uniforms['proj'])
			# render on self.context
			self.reflectmap.use(0)
			self.va.render(mgl.TRIANGLES)
	
	def identify(self, view):
		if self.va:
			self.ident_shader['layer'] = self.layer
			self.ident_shader['start_ident'] = view.identstep(self.vertices.nident)
			self.ident_shader['view'].write(view.uniforms['view'] * self.vertices.world)
			self.ident_shader['proj'].write(view.uniforms['proj'])
			# render on self.context
			self.va_ident.render(mgl.TRIANGLES)

class GhostDisplay:
	def __init__(self, scene, vertices, normals, faces, color, layer=0):
		self.color = color
		self.layer = layer
		self.vertices = vertices
		
		# load the shader
		def load(scene):
			shader = scene.ctx.program(
						vertex_shader=open(ressourcedir+'/shaders/solid.vert').read(),
						fragment_shader=open(ressourcedir+'/shaders/ghost.frag').read(),
						)
			# setup some uniforms
			return shader
		self.shader = scene.ressource('shader_ghost', load)
		self.ident_shader = scene.ressource('shader_subident')
		# allocate buffers
		if faces is not None and len(faces) and vertices.vb_positions:
			self.vb_faces = scene.ctx.buffer(np.array(faces, 'u4', copy=False))
			self.vb_normals = scene.ctx.buffer(np.array(normals, 'f4', copy=False))
			self.va = scene.ctx.vertex_array(
					self.shader, 
					[	(vertices.vb_positions, '3f', 'v_position'), 
						(self.vb_normals, '3f', 'v_normal'),
						(vertices.vb_flags, 'u1', 'v_flags')],
					self.vb_faces,
					)
			
			self.va_ident = scene.ctx.vertex_array(
					self.ident_shader, 
					[	(vertices.vb_positions, '3f', 'v_position'),
						(vertices.vb_idents, 'u2', 'item_ident')], 
					self.vb_faces,
					)
		else:
			self.va = None
			
	def __del__(self):
		if self.va:
			self.va.release()
			self.va_ident.release()
			self.vb_faces.release()
			self.vb_normals.release()
	
	def render(self, view):
		if self.va:
			# setup uniforms
			self.shader['select_color'].write(settings.display['select_color_line'])
			self.shader['normal_color'].write(self.color)
			self.shader['world'].write(self.vertices.world)
			self.shader['view'].write(view.uniforms['view'])
			self.shader['proj'].write(view.uniforms['proj'])
			self.shader['layer'] = self.layer
			view.scene.ctx.disable(mgl.DEPTH_TEST)
			# render on self.context
			self.va.render(mgl.TRIANGLES)
			view.scene.ctx.enable(mgl.DEPTH_TEST)
	
	def identify(self, view):
		if self.va:
			self.ident_shader['start_ident'] = view.identstep(self.vertices.nident)
			self.ident_shader['view'].write(view.uniforms['view'] * self.vertices.world)
			self.ident_shader['proj'].write(view.uniforms['proj'])
			self.ident_shader['layer'] = self.layer
			# render on self.context
			self.va_ident.render(mgl.TRIANGLES)
	
class LinesDisplay:
	def __init__(self, scene, vertices, lines, color, alpha=1, layer=0):
		self.layer = layer
		self.color = fvec4(fvec3(color), alpha)
		self.select_color = fvec4(settings.display['select_color_line'], alpha)
		self.vertices = vertices
		
		# load the line shader
		self.shader = scene.ressource('shader_wire', shader_wire)
		self.ident_shader = scene.ressource('shader_subident')
		if lines is not None and len(lines) and vertices.vb_positions:
			# allocate buffers
			self.vb_lines = scene.ctx.buffer(np.array(lines, dtype='u4', copy=False))
			self.va = scene.ctx.vertex_array(
						self.shader,
						[	(vertices.vb_positions, '3f', 'v_position'),
							(vertices.vb_flags, 'u1', 'v_flags')],
						self.vb_lines,
						)
			self.va_ident = scene.ctx.vertex_array(
					self.ident_shader, 
					[	(vertices.vb_positions, '3f', 'v_position'),
						(vertices.vb_idents, 'u2', 'item_ident')], 
					self.vb_lines,
					)
		else:
			self.va = None
			
	def __del__(self):
		if self.va:
			self.va.release()
			self.va_ident.release()
			self.vb_lines.release()
	
	def render(self, view):
		if self.va:
			self.shader['color'].write(self.color)
			self.shader['select_color'].write(self.select_color)
			self.shader['view'].write(view.uniforms['view'] * self.vertices.world)
			self.shader['proj'].write(view.uniforms['proj'])
			self.shader['layer'] = self.layer
			self.va.render(mgl.LINES)
		
	def identify(self, view):
		if self.va:
			self.ident_shader['start_ident'] = view.identstep(self.vertices.nident)
			self.ident_shader['view'].write(view.uniforms['view'] * self.vertices.world)
			self.ident_shader['proj'].write(view.uniforms['proj'])
			self.ident_shader['layer'] = self.layer
			self.va_ident.render(mgl.LINES)
		
class PointsDisplay:
	def __init__(self, scene, vertices, indices=None, color=None, ptsize=3, layer=0):
		self.color = fvec4(color or settings.display['point_color'], 1)
		self.select_color = fvec4(settings.display['select_color_line'], 1)
		self.ptsize = ptsize
		self.layer = layer
		self.vertices = vertices
		
		# load the line shader
		self.shader = scene.ressource('shader_wire', shader_wire)
		self.ident_shader = scene.ressource('shader_subident')
		# allocate GPU objects
		if indices is not None and len(indices) and vertices.vb_positions:
			self.vb_indices = scene.ctx.buffer(np.array(indices, dtype='u4', copy=False))
			self.va = scene.ctx.vertex_array(
						self.shader,
						[	(vertices.vb_positions, '3f', 'v_position'),
							(vertices.vb_flags, 'u1', 'v_flags')],
						self.vb_indices,
						)
			self.va_ident = scene.ctx.vertex_array(
					self.ident_shader, 
					[	(vertices.vb_positions, '3f', 'v_position'),
						(vertices.vb_idents, 'u2', 'item_ident')], 
					self.vb_indices,
					)
		else:
			self.va = None
			
	def __del__(self):
		if self.va:
			self.va.release()
			self.va_ident.release()
			self.vb_indices.release()
	
	def render(self, view):
		if self.va:
			self.shader['layer'] = self.layer
			self.shader['color'].write(self.color)
			self.shader['select_color'].write(self.select_color)
			self.shader['view'].write(view.uniforms['view'] * self.vertices.world)
			self.shader['proj'].write(view.uniforms['proj'])
			view.scene.ctx.point_size = self.ptsize
			self.va.render(mgl.POINTS)
	
	def identify(self, view):
		if self.va:
			scene.subident_shader['layer'] = self.layer
			scene.subident_shader['start_ident'] = view.identstep(self.vertices.nident)
			scene.subident_shader['view'].write(view.uniforms['view'] * self.vertices.world)
			scene.subident_shader['proj'].write(view.uniforms['proj'])
			view.ctx.point_size = self.ptsize
			self.va_ident.render(mgl.POINTS)

def digitfit(n):
	''' return the number of zeros in the digital representation of n '''
	s = 0
	while n:
		if not n%10:	s += 1
		n //=10
	return s

class GridDisplay(Display):
	''' display a grid clipped to the view, helping to appreciate distances
		The grid display distances in the plane of the `center` point
		
		:unit:      factor on the distance between dots
		:color:     base color of the grid, with an alpha channel
		:contrast:  contrast between the smallest division and the biggest
	'''
	def __init__(self, scene, center, unit=1, color=None, contrast=1.8):
		def load(scene):
			n = 101
			h = n//2
			pts = np.empty((n,n), dtype='f4, f4, f2')
			for i in range(n):
				for j in range(n):
					pts[i,j] = (i-h, j-h, (1+min(digitfit(i), digitfit(j))))
			
			shader = scene.ctx.program(
						vertex_shader=open(ressourcedir+'/shaders/viewgrid.vert').read(),
						fragment_shader=open(ressourcedir+'/shaders/viewgrid.frag').read(),
						)
			vb = scene.ctx.buffer(pts)
			va = scene.ctx.vertex_array(shader, [(vb, '2f4 f2', 'v_position', 'v_opacity')])
			return shader, va
		self.shader, self.va = scene.ressource('viewgrid', load)
		self.unit = unit
		self.center = fvec3(center or 0)
		self.color = fvec4(color) if color else fvec4(settings.display['point_color'],1)
		self.contrast = contrast
	
	def render(self, view):
		center = fvec3( view.uniforms['view'] * self.world * fvec4(self.center,1) )
		if not center.z < 0:	return
		zlog = log(-center.z)/log(10)
		sizelog = floor(zlog)
		
		view.scene.ctx.point_size = 1/400 * view.fb_screen.height
		self.shader['color'].write(fvec4(
				self.color.rgb, 
				self.color.a * exp(self.contrast*(-1+sizelog-zlog)),
				))
		self.shader['size'] = self.unit * 10**(sizelog-1)
		self.shader['centerdist'] = -center.z
		self.shader['proj'].write(view.uniforms['proj'])
		self.shader['contrast'] = self.contrast
		self.va.render(mgl.POINTS)
	
	def stack(self, scene):
		return ((), 'screen', 3, self.render),
		

class SplineDisplay(Display):
	''' display for spline curve, with handles around'''
	def __init__(self, scene, handles, curve, color=None):
		self.color = color or fvec4(settings.display['line_color'], 1)
		self.color_handles = fvec4(settings.display['annotation_color'], 0.6)
		self.box = npboundingbox(handles)
		ctx = scene.ctx
		self.vb_handles = ctx.buffer(handles)
		self.vb_curve = ctx.buffer(curve)
		
		self.shader = scene.ressource('shader_uniformcolor', shader_uniformcolor)
		self.va_handles = ctx.vertex_array(self.shader, [(self.vb_handles, '3f4', 'v_position')])
		self.va_curve = ctx.vertex_array(self.shader, [(self.vb_curve, '3f4', 'v_position')])
		
		self.shader_ident = scene.ressource('shader_ident')
		self.va_ident = ctx.vertex_array(self.shader_ident, [(self.vb_curve, '3f4', 'v_position')])
		
	def __del__(self):
		self.va_handles.release()
		self.va_curve.release()
		self.va_ident.release()
		self.vb_handles.release()
		self.vb_curve.release()
		
	def render(self, view):
		self.shader['view'].write(view.uniforms['view'] * self.world)
		self.shader['proj'].write(view.uniforms['proj'])
		view.scene.ctx.point_size = 4
		
		self.shader['layer'] = -2e-6
		self.shader['color'].write(self.color_handles if not self.selected else fvec4(settings.display['select_color_line'],self.color_handles[3]))
		self.va_handles.render(mgl.POINTS)
		self.va_handles.render(mgl.LINE_STRIP)
		
		self.shader['layer'] = -1e-6
		self.shader['color'].write(self.color if not self.selected else fvec4(settings.display['select_color_line'],self.color[3]))
		self.va_curve.render(mgl.LINE_STRIP)
		
	def identify(self, view):
		self.shader_ident['ident'] = view.identstep(1)
		self.shader_ident['view'].write(view.uniforms['view'] * self.world)
		self.shader_ident['proj'].write(view.uniforms['proj'])
		self.va_ident.render(mgl.LINE_STRIP)
		
	def stack(self, scene):
		return (	((), 'screen', 1, self.render),
					((), 'ident', 1, self.identify)	)
					
					
					
class VoxelDisplay(Display):
	''' display a voxel as a scalar field.
		its opacity is computed using its density as absorbance.
		
		Parameters:
			
			voxel:	a numpy array with dimension 3, no specific value range is required
			
		Attributes:
		
			space:	transformation matrix from voxel normalized coordinates (position in the array as range 0-1) to local coordinates
			world:	transformation matrix from local to world coordinates
			value_range:	
				
				tuple `(min, max)`  giving the min and max values to display
				- values below min will be transparent
				- values above max will be fully opaq
				
			color_range:	tuple `(fvec4, fvec4)` of colors matching the value range bounds
		
		value examples:
		
			* value_range
			
				```
				(0.4, 0.9)  # erase the lower value and saturate the maxium ones
				```
				
			* color_range
			
				```
				(fvec4(0,0,1,1), fvec4(0,1,0,1))	 # blue-green
				(fvec4(0,0.5,1,1), fvec4(1,1,0,1))   # lightblue-yellow
				(fvec4(1,0,0.3,1), fvec4(1,1,0,1))   # red-yellow
				(fvec4(0,0,1,1), fvec4(1,0.5,0.1,1)) # blue-orange
				(fvec4(0,0,1,1), fvec4(1,0.1,0.1,1)) # blue-red
				(fvec4(0,0,0,1), fvec4(1,1,1,1))     # black-white
				(fvec4(1,1,1,1), fvec4(1,1,1,1))     # white-white
				```
	'''
	def __init__(self, scene, voxel: 'ndarray', space: 'fmat4', 
				value=(0, 1), 
				color=(fvec4(0,0,1,1), fvec4(0,1,0,1))):
		
		self.voxel = scene.ctx.texture3d(
			voxel.shape, 
			1, 
			voxel.transpose(2,1,0).copy().astype('f2'),  # axis permutation for opengl
			dtype='f2',
			)
		self.voxel.repeat_x = False
		self.voxel.repeat_y = False
		self.voxel.repeat_z = False
		self.space = space
		self.value_range = value
		self.color_range = color

		def load(scene):
			from . import generation
			from .mesh import glmarray
			
			# load shader
			shader = scene.ctx.program(
						vertex_shader=open(ressourcedir+'/shaders/voxel.vert').read(),
						fragment_shader=open(ressourcedir+'/shaders/voxel.frag').read(),
						)
			# load vertex buffer for the brick
			brick = generation.brick(min=vec3(0), max=vec3(1)) .flip()
			pts = []
			for face in brick.faces:
				pts.extend(brick.facepoints(face))
			vb = scene.ctx.buffer(glmarray(pts))
			
			return shader, vb
		
		self.shader, self.vb = scene.ressource('shader_voxel', load)
		self.va = scene.ctx.vertex_array(
				self.shader,
				[(self.vb, '3f', 'v_position')],
				)
				
	def __del__(self):
		self.va.release()
		self.voxel.release()

	def render(self, view):
		view.scene.ctx.enable(mgl.DEPTH_TEST | mgl.CULL_FACE)
		self.shader['value_min'] = self.value_range[0]
		self.shader['value_max'] = self.value_range[1]
		self.shader['color_min'].write(self.color_range[0])
		self.shader['color_max'].write(self.color_range[1])
		self.shader['view'].write(view.uniforms['view'] * self.world * self.space)
		self.shader['proj'].write(view.uniforms['proj'])
		self.voxel.use(0)
		self.va.render(mgl.vertex_array.TRIANGLES)
		
	def stack(self, scene):
		return ((), 'screen', 10, self.render),




def tupledisplay(scene, t):
	if primitives.isaxis(t):	return AxisDisplay(scene, t)
	# if not found: empty display
	return Display()

overrides.update({
	vec3:   PointDisplay,
	tuple:  tupledisplay,
	Box:    BoxDisplay,
	})

	



'''	 The settings module holds dictionnaries each aspect of the madcad library.

dictionnaries:
	:primitives:  default settings for mesh and primitive operations
	:display:	visual settings to display the 3D objects
	:scene:		for what and how to display in the 3D scene
	:controls:	preferences for the controls of the Scene widget
'''

from math import pi, ceil, floor, sqrt
from glm import fvec3, fvec4
import sys, os, yaml
from os.path import dirname, exists


display = {
	'field_of_view': pi/6,
	'view_font_size': 8,
	'line_width': 1.,
	
	'background_color': fvec3(0,0,0),
	'select_color_face': fvec3(0.01, 0.05, 0.03),
	'select_color_line': fvec3(0.5, 1, 0.6),
	'highlight_color': fvec3(0.1, 0.2, 0.2),
	
	'solid_color': fvec3(0.2, 0.2, 0.2),
	'solid_color_front': 1.,
	'solid_color_side': 0.2,
	'solid_reflectivity': 6,
	'solid_reflect': 'skybox-white.png',
	'line_color': fvec3(0.9, 0.9, 0.9),
	'point_color': fvec3(0.9, 0.9, 0.9),
	'schematics_color': fvec3(0.3, 0.8, 1),
	# According to moderngl docs changing this is not guaranteed to work
	# (depends on OpenGL implementation)
	'line_width': 1.0,
	'solver_error_color': fvec3(1, 0.3, 0.2),
	'annotation_color': fvec3(0.2, 0.7, 1),
	'sharp_angle': pi/6,
	
	'system_theme': True,
	}

scene = {
	'projection': 'Perspective',
	
	'display_faces': True,
	'display_groups': True,
	'display_points': False,
	'display_wire': False,
	'display_grid': True,
	'display_annotations': True,
	'surface_shading': True,
	
	'lock_solids': True,
	
	'debug_points': False,
	'debug_faces': False,
	'debug_groups': False,
	}
		
controls = {
	'navigation': 'Turntable',
	'snap_dist': 10,	# pixel distance to click items
	}

primitives = {
	'curve_resolution': ['rad', pi/16],	# angle maximal pour discretisation des courbes
	}

# get configuration directory depending on OS
if sys.platform == 'win32':
	home = os.getenv('USERPROFILE')
	configdir = home+'/AppData/Local'
else:
	home = os.getenv('HOME')
	configdir = home+'/.config'

config = configdir+'/madcad/pymadcad.yaml'
settings = {'display':display, 'scene':scene, 'controls':controls, 'primitives':primitives}


def install():
	''' create and fill the config directory if not already existing '''
	if not exists(config):
		os.makedirs(dirname(config), exist_ok=True)
		dump()
		
def clean():
	''' delete the default configuration file '''
	os.rm(config)

def load(file=None):
	''' load the settings directly in this module, from the specified file or the default one '''
	if not file:	file = config
	if isinstance(file, str):	file = open(file, 'r')
	changes = yaml.safe_load(file)
	def update(dst, src):
		for key in dst:
			if key in src:
				if isinstance(dst[key], dict) and isinstance(src[key], dict):	
					update(dst[key], src[key])
				elif isinstance(dst[key], fvec3):	dst[key] = fvec3(src[key])
				elif isinstance(dst[key], fvec4):	dst[key] = fvec4(src[key])
				else:
					dst[key] = src[key]
	update(settings, changes)

def dump(file=None):
	''' load the current settings into the specified file or to the default one '''
	if not file:	file = config
	if isinstance(file, str):	file = open(file, 'w')
	yaml.add_representer(fvec3, lambda dumper, data: dumper.represent_list(round(f,3) for f in data))
	yaml.add_representer(fvec4, lambda dumper, data: dumper.represent_list(round(f,3) for f in data))
	file.write(yaml.dump(settings, default_flow_style=None, width=40, indent=4))




def curve_resolution(length, angle, param=None):
	''' return the subdivision number for a curve, using the given or setting specification

		:length:  is the curvilign length of the curve
		:angle:   is the integral of the absolute curvature (total rotation angle)
	'''
	kind, prec = param or primitives['curve_resolution']
	if kind == 'div':
		res = prec
	elif kind == 'm':
		res = ceil(length / prec)
	elif kind == 'rad':
		res = floor(angle / prec)
	elif kind == 'radm':
		res = floor(length*angle / prec)
	elif kind == 'sqradm':
		res = floor(sqrt(length*angle) / prec)
	else:
		raise ValueError("unknown type for round_limit: {}".format(repr(kind)))
	res = max(res, floor(angle * 2/pi))
	return res

def getparam(levels: list, key):
	''' get the first found value for key through the given dictionnaries.
		Dictionnaries are tested successively until the matching value is found. If no value is found, None is returned
	'''
	for d in levels:
		if d is not None:
			if key in d:	return d[key]
	return None

def use_qt_colors():
	''' set the color settings to fit the current system colors '''
	from .mathutils import fvec3, mix, distance
	from PyQt5.QtWidgets import QApplication
	palette = QApplication.instance().palette()
	def qtc(role):
		''' convert a QColor or QPalette role to fvec3'''
		c = palette.color(role)
		return fvec3(c.red(), c.green(), c.blue()) / 255
		
	
	selection = mix(fvec3(0.4, 1, 0), qtc(palette.Highlight), 0.6)
	selection *= mix(1/max(selection), max(qtc(palette.Text)), 0.3)
	display.update({
		'background_color': qtc(palette.Base),
		'select_color_face': selection * 0.05,
		'select_color_line': selection * 1.1,
		'line_color': qtc(palette.Text),
		'point_color': qtc(palette.Text),
		'solid_color': mix(qtc(palette.Text), qtc(palette.Window), 0.7),
		'schematics_color': qtc(palette.Link),
		'annotation_color': mix(qtc(palette.Highlight), qtc(palette.Text), 0.5),
		})


# automatically load settings in the file exist
try:	load()
except FileNotFoundError:	pass





class Asso(object):
	''' Assoassociative container.
		This is a sort dict that stores as many values as we want for each key. The return value for a key is then always an iterable of the associated values, even when the value set to the key is unique and is not an iterator.
		
		Asso(iterable)				the iterable must yields (key,value) pairs
		Asso(keys=[], values=[])	build form existing lists, assuming they are already sorted
	'''
	__slots__ = '_table'
	
	def __init__(self, iter=None):
		if isinstance(iter, Asso):
			self._table = dict(iter.table)
		else:
			self._table = {}
			if iter:
				self.update(iter)
		
	def __getitem__(self, key):
		''' return an iterable of the objects associated to the given key '''
		i = 0
		while True:
			k = (i,key)
			if k not in self._table:
				break
			yield self._table[k]
			i += 1
			
	def add(self, key, value):
		''' associate a new value to this key '''
		i = 0
		while True:
			k = (i,key)
			if k not in self._table:
				self._table[k] = value
				return
			i += 1
	
	def remove(self, key, value):
		l = self.connexity(key)-1
		i = l
		while i>=0:
			k = (i,key)
			if self._table[k] == value:
				last = (l,key)
				self._table[k] = self._table[last]
				del self._table[last]
				return
			i -= 1
		raise KeyError('key not in table')
		
	def discard(self, key, value):
		''' remove all (key,value) pair of that table '''
		l = self.connexity(key)-1
		i = l
		while i>=0:
			k = (i,key)
			if self._table[k] == value:
				last = (l,key)
				self._table[k] = self._table[last]
				del self._table[last]
				l -= 1
			i -= 1
	
	def update(self, other):
		''' append all key,value associations to this Asso '''
		if isinstance(other, dict):
			other = other.items()
		for k,v in other:
			self.add(k,v)
		
	def __add__(self, other):
		new = Asso()
		new._table = dict(self._table)
		new.update(other)
		return new
		
		
	def __delitem__(self, key):
		''' delete all associations with this key '''
		i = 0
		while True:
			k = (i,k)
			if k not in self._table:
				break
			del self._table[k]
			i += 1
	
	def clear(self):
		''' empty the container '''
		self._table.clear()

	def items(self):
		''' iterator of (key, value) pairs '''
		for k,v in self._table.items():
			yield k[1],v

	def keys(self):
		''' iterator of the keys '''
		for k,v in self._table:
			yield k[1]
		
	def values(self):
		''' iterator of the values '''
		return self._table.values()
		
	def connexity(self, key):
		''' return the number of values associated to the given key '''
		i = 0
		while True:
			k = (i,key)
			if k not in self._table:
				break
			i += 1
		return i
		

	def __contains__(self, key):
		''' return True if key is associated to something '''
		return (0,key) in self._table
		
	def __repr__(self):
		return 'Asso([{}])'.format(', '.join(
					'({}, {})'.format(repr(k), repr(v))	
						for k,v in self.items()
					))

if __name__ == '__main__':
	from nprint import nprint

	m = Asso([
			(1,'truc'), 
			(2,'machin'),
			(12, 'chose'),
			(1, 'bidule'),
			])
	nprint('created', m)
	
	m.update([
		(1,'machin'),
		(-1,'gna'),
		])
	nprint('updated', m)
	
	m.add(1, 2)
	nprint('inserted', m)
	
	assert set(m[1]) == {'truc', 'bidule', 2, 'machin'}
	assert set(m[12]) == {'chose'}





from mathutils import *

class Derived:
	__slots__ = 'value', 'der', 'vars'
	
	def __init__(self, value, derivatives=None, vars=None):
		self.value = value
		if derivatives is None and vars is None:
			self.der = [1]
			vars = [self.value]
		else:
			self.derivatives = derivatives
			self.vars = vars
	
	def __add__(self, other):
		return Derived(self.value+other.value, self.grad+other.grad)
	def __mul__(self, other):
		return Derived(self.value*other.value, self.grad*other.value + other.grad*self.value)

def compose(func):
	dfunc = derivatives[func]
	def func(value):
		return Derived(func(value.value), value.grad * dfunc(value.value))
	return func

def dlength(v):
	indev



# This file is part of pymadcad,  distributed under license LGPL v3

from .mathutils import *
from .primitives import isaxis
from .kinematic import Screw, WireDisplay, Scheme, Joint
from .mesh import Mesh, Wire, web, Web
from . import generation, primitives

__all__ = ['Pivot', 'Plane', 'Track', 'Gliding', 'Ball', 'Punctiform', 'Gear', 'Helicoid']

'''
TODO:
	Ring		linéaire annulaire
	Hinge		linéaire
	Rack		crémaillere
	
	Rack(onesolid, onceagain, (O,z), x, 1, offset=0),	# cremaillere
'''

	
class InSolid:	# TODO test
	def __init__(self, solid, points, reference=None):
		self.solid = solid
		self.points = points
		self.reference = reference or [p*solid.transform() for p in points]	# if no reference position is provided, then we use the current position in the solid
		# remove origin from references
		center = sum(reference)/len(reference)
		for p in self.reference:	p -= center
	
	slvvars = 'solid', 'points',

	def pose(self):
		center = sum(self.points)/len(self.points)
		rot = vec3(0)
		for pt,ref in zip(self.points, self.reference):
			v, r = normalize(pt-ref), normalize(ref)
			c = cross(v,r)
			if not glm.any(isnan(c)):
				angle = acos(dot(v,r))
				rot += angleAxis(angle, c/sin(angle))
		rot = normalize(rot)
		return self.solid.position+center, self.solid.orientation+rot

cornersize = 0.1

def solidtransform_axis(solid, obj):
	rot = solid.orientation
	return (rot*obj[0] + solid.position, rot*obj[1])
def solidtransform_base(solid, obj):
	rot = solid.orientation
	if len(obj) == 3:	return (rot*obj[0] + solid.position, rot*obj[1], rot*obj[2])
	if len(obj) == 4:	return (rot*obj[0] + solid.position, rot*obj[1], rot*obj[2], rot*obj[3])
	
	
class Welded(Joint):
	def __init__(self, s1, s2, transform=None):
		self.solids = (s1, s2)
		self.transform = transform or affineInverse(s1.pose) * s2.pose
		self.position = (vec3(0), vec3(0))
		
	def corrections(self):
		p1 = dmat4x3(self.solids[0].pose * self.transform)
		p2 = dmat4x3(self.solids[1].pose)
		t = p2[3] - p1[3]
		r = cross(p1[0], p2[0]) + cross(p1[1], p2[1]) + cross(p1[2], p2[2])
		return Screw(t,r,p1[3]), Screw(-t,-r,p2[3])
		
	def scheme(self, solid, size, junc=None) -> Scheme:
		return Scheme([], [], [], [])

class Pivot(Joint):
	''' Junction for rotation only around an axis
	
		classical definition:	Pivot (axis)
		no additional information is required by the initial state
		this class holds an axis for each side
	'''
	def __init__(self, s1, s2, a1, a2=None, position=None):
		self.solids = (s1,s2)
		self.axis = (a1, a2 or a1)
		self.position = position or (self.axis[0][0], self.axis[1][0])
	
	slvvars = 'solids',
	def fit(self):
		a0,a1 = solidtransform_axis(self.solids[0], self.axis[0]), solidtransform_axis(self.solids[1], self.axis[1])
		return distance(a0[0],a1[0])**2 + 1-dot(a0[1],a1[1])
	
	def corrections(self):
		a0,a1 = solidtransform_axis(self.solids[0], self.axis[0]), solidtransform_axis(self.solids[1], self.axis[1])
		r = cross(a0[1], a1[1])
		t = a1[0] - a0[0]
		#return Screw(r,t,a0[0]), Screw(-r,-t,a1[0]) 	# velocity torsor
		return Screw(t,r,a0[0]), Screw(-t,-r,a1[0]) 	# force torsor
	
	def transmitable(self, action):
		axis = solidtransform_axis(self.solids[0], self.axis[0])
		normal = normalize(a0[1] + a1[1])
		return Screw(action.resulting, noproject(action.momentum, normal), action.position)
		
	def scheme(self, solid, size, junc=None) -> Scheme:
		''' return a Scheme to render the junction from the given solid side
			size is the desired size of the junction
			junc is the point the junction is linked with the scheme by
		'''
		if solid is self.solids[0]:
			radius = size/4
			axis = self.axis[0]
			center = axis[0] + project(self.position[0]-axis[0], axis[1])
			cyl = generation.extrusion(
						axis[1]*size * 0.8, 
						primitives.Circle(
								(center-axis[1]*size*0.4, axis[1]), 
								size/4, 
								resolution=('div', 16),
						))
			sch = Scheme(cyl.points, cyl.faces, [], list(cyl.outlines_unoriented()))
			if junc:
				v = normalize(noproject(junc - center, axis[1]))
				if not isfinite(v):
					v,_,_ = dirbase(axis[1])
				p = center + v*size/4
				sch.extend(Scheme(
							[p, p + axis[1]*size*cornersize, p + v*size*cornersize, junc],
							[], 
							[(0,1,2)],
							[(0,2), (2,3)],
							))
			return sch
					
			
			#axis = self.axis[0]
			#center = axis[0] + project(self.position[0]-axis[0], axis[1])
			#cyl = generation.extrusion(
						#axis[1]*size * 0.8, 
						#primitives.Circle(
								#(center-axis[1]*size*0.4, axis[1]), 
								#size/4, 
								#resolution=('div', 16),
						#))
			
			#sch = (scheme.Scheme(space=scheme.world, color=fvec4(settings.display['schematics_color'],1))
					#.add(cyl, shader='ghost')
					#.add(cyl.outlines(), shader='line')
					#)
			
			#v = junc - center
			#v = normalize(noproject(v,axis[1]))
			#if isfinite(v):
				#x,_,_ = dirbase(axis[1])
			#p = center + v*size/4
			#sch.add(
				#Mesh([p, p+axis[1]*size*cornersize, p + x*size*cornersize], [(0,1,2)]),
				#shader='fill')
			#sch.add([p, junc], shader='line')
				
			#return sch
					
		elif solid is self.solids[1]:
			radius = size/4
			axis = self.axis[1]
			center = axis[0] + project(self.position[1]-axis[0], axis[1])
			side = axis[1] * size * 0.5
			if dot(junc-axis[0], axis[1]) < 0:
				side = -side
			if dot(junc-center-side, side) < 0:
				attach = side + normalize(noproject(junc-center, axis[1]))*radius
			else:
				attach = side
			c1 = primitives.Circle(
							(center-axis[1]*size*0.5, -axis[1]), 
							radius, 
							resolution=('div', 16),
							).mesh()
			c2 = primitives.Circle(
							(center+axis[1]*size*0.5, axis[1]), 
							radius, 
							resolution=('div', 16),
							).mesh()
			s1 = generation.flatsurface(c1)
			s2 = generation.flatsurface(c2)
			l = len(c1.points)
			indices = [(i-1,i)  for i in range(1,l-1)]
			indices.append((l-1,0))
			
			s = Scheme([center-side, center+side, center+attach, junc], [], [], [(0,1), (2,3)])
			s.extend(Scheme(c1.points, s1.faces, [], c1.edges()))
			s.extend(Scheme(c2.points, s2.faces, [], c2.edges()))
			return s

class Planar(Joint):
	''' Joint for translation in 2 directions and rotation around the third direction 
		
		classical definition:	Planar (direction vector)
		the initial state requires an additional distance between the solids
		this class holds an axis for each side, the axis origins are constrained to share the same projections on the normal
	'''
	def __init__(self, s1, s2, a1, a2=None, position=None):
		self.solids = (s1, s2)
		self.axis = (a1, a2 or a1)
		self.position = position or (self.axis[0][0], self.axis[1][0])
	
	slvvars = 'solids',
	def fit(self):
		a0,a1 = solidtransform_axis(self.solids[0], self.axis[0]), solidtransform_axis(self.solids[1], self.axis[1])
		normal = normalize(a0[1] + a1[1])
		return dot(a0[0]-a1[0], normal) **2 - dot(a0[1], a1[1])
	
	def corrections(self):
		a0,a1 = solidtransform_axis(self.solids[0], self.axis[0]), solidtransform_axis(self.solids[1], self.axis[1])
		r = cross(a0[1], a1[1])
		t = project(a1[0] - a0[0], normalize(a0[1] + a1[1]))
		
		#return Screw(r,t,a0[0]), Screw(-r,-t,a1[0]) 	# velocity torsor
		return Screw(t,r,a0[0]), Screw(-t,-r,a1[0]) 	# force torsor
	
	def transmitable(self, action):
		a0,a1 = solidtransform_axis(self.solids[0], self.axis[0])
		normal = normalize(a0[1] + a1[1])
		return Screw(project(action.resulting, normal), noproject(action.momentum, normal), action.position)
	
		
	
	def scheme(self, solid, size, junc=None):
		if   solid is self.solids[0]:		(center, normal), position = self.axis[0], self.position[0]
		elif solid is self.solids[1]:		(center, normal), position = self.axis[1], self.position[1]
		else:	return
		center = position - project(position - center, normal)
		if dot(junc-center, normal) < 0:	normal = -normal
		if solid is self.solids[1]:			normal = -normal
		x,y,z = dirbase(normal)
		c = center + size*0.1 * z
		x *= size*0.7
		y *= size*0.7
		return Scheme(
			[c+(x+y), c+(-x+y), c+(-x-y), c+(x-y), c, c+cornersize*x, c+cornersize*z, junc],
			[(0,1,2), (0,2,3)],
			[(4,5,6)],
			[(0,1),(1,2),(2,3),(3,0),(4,6),(6,7)],
			)

class Track(Joint):
	''' Joint for translation only in a direction 
		
		classical definition:  Track (direction vector)
		the initial state requires more parameters: the relative placements of the solids
		this class holds a base for each of the solids, bases are (0,X,Y)  where X,Y are constrained to keep the same direction across bases, and the bases origins lays on their common Z axis
	'''
	def __init__(self, s1, s2, b1, b2=None, position=None):
		self.solids = (s1, s2)
		self.bases = (b1, b2 or b1)
		self.position = (self.bases[0][0], self.bases[1][0])
	
	slvvars = 'solids',
	def fit(self):
		b0, b1 = solidtransform_base(self.solids[0], self.bases[0]), solidtransform_base(self.solids[1], self.bases[1])
		t = b1[0] - b0[0]
		x = normalize(b0[1]+b1[1])
		y = normalize(b0[2]+b1[2])
		return (	1-dot(cross(b0[1],b0[2]), cross(b1[1],b1[2]))
				+	dot(t,x) **2
				+	dot(t,y) **2
				)
	
	def corrections(self):
		b0, b1 = solidtransform_base(self.solids[0], self.bases[0]), solidtransform_base(self.solids[1], self.bases[1])
		z0, z1 = cross(b0[1],b0[2]), cross(b1[1],b1[2])
		r = (	cross(z0,z1)
			+	cross(b0[1], b1[1])
			+	cross(b0[2], b1[2])
			)
		t = b1[0] - b0[0]
		return (
			Screw(noproject( t, z1),  r,b0[0]), 
			Screw(noproject(-t, z0), -r,b1[0]),
			)
	
	def transmitable(self, action):
		z0, z1 = cross(b0[1],b0[2]), cross(b1[1],b1[2])
		normal = normalize(z0 + z1)
		return Screw(noproject(action.resulting, normal), action.momentum, action.position)
	
	def scheme(self, solid, size, junc):
		if solid is self.solids[0]:
			o,x,y = self.bases[0]
			y = normalize(noproject(y,x))
			z = normalize(cross(x,y))
			s = 0.25*size
			line = Web(
						[(x-y)*s, (x+y)*s, (x+y)*s, (-x+y)*s, (-x+y)*s, (-x-y)*s, (-x-y)*s, (x-y)*s],
						[(0,1),(2,3),(4,5),(6,7)],
						)
			line.transform(o-size/2*z)
			ext = generation.extrusion(size*z, line)
			l = len(ext.points)
			v = junc - o
			if abs(dot(v,x)) > abs(dot(v,y)):
				v = x if dot(v,x)>0 else -x
			else:
				v = y if dot(v,y)>0 else -y
			p = o + v*0.25*size
			ext.points.append(p)
			ext.points.append(p + z*size*cornersize)
			ext.points.append(p + v*size*cornersize)
			ext.points.append(junc)
			return Scheme(
					ext.points, 
					ext.faces, 
					[(l,l+1,l+2)], 
					[	(l+2, l+3),
						(0,1),(2,3),(4,5),(6,7),
						(0,8), (2,10), (4,12), (6,14), 
						(8,9),(10,11),(12,13),(14,15)],
					)
		
		elif solid is self.solids[1]:
			o,x,y = self.bases[1]
			y = noproject(y,x)
			z = cross(x,y)
			s = 0.15*size
			line = Web(
						[(x+y)*s, (-x+y)*s, (-x-y)*s, (x-y)*s],
						[(0,2),(1,3)],
						)
			line.transform(o-size/2*z)
			ext = generation.extrusion(size*z, line)
			l = len(ext.points)
			v = junc - o
			v = z if dot(v, z) > 0 else -z
			p = o + v*size/2
			ext.points.append(p)
			ext.points.append(junc)
			return Scheme(
					ext.points, 
					[], 
					[], 
					[	(l, l+1),
						(0,2), (1,3),
						(0,4), (1,5), (2,6), (3,7),
						(4,6), (5,7)],
					)

class Gliding(Joint):
	''' Joint for rotation and translation around an axis 
		
		classical definition:  Gliding pivot (axis)
		the initial state doesn't require more data
		this class holds an axis for each side
	'''
	def __init__(self, s1, s2, a1, a2=None, position=None):
		self.solids = (s1,s2)
		self.axis = (a1, a2 or a1)
		self.position = position or (self.axis[0][0], self.axis[1][0])
	
	slvvars = 'solids',
	def fit(self):
		a0,a1 = solidtransform_axis(self.solids[0], self.axis[0]), solidtransform_axis(self.solids[1], self.axis[1])
		return distance(a0[0],a1[0])**2 + 1-dot(a0[1],a1[1])
	
	def corrections(self):
		a0,a1 = solidtransform_axis(self.solids[0], self.axis[0]), solidtransform_axis(self.solids[1], self.axis[1])
		r = cross(a0[1], a1[1])
		t = a1[0] - a0[0]
		return (
				Screw(noproject( t,a0[1]),  r,a0[0]), 
				Screw(noproject(-t,a1[1]), -r,a1[0]),
				) # force torsor
	
	def transmitable(self, action):
		axis = solidtransform_axis(self.solids[0], self.axis[0])
		normal = normalize(a0[1] + a1[1])
		return Screw(
				noproject(action.resulting, normal), 
				noproject(action.momentum, normal), 
				action.position)
		
	def scheme(self, solid, size, junc):
		''' return primitives to render the junction from the given solid side
			size is the desired size of the junction
			junc is the point the junction is linked with the scheme by
		'''
		if solid is self.solids[0]:
			radius = size/4
			axis = self.axis[0]
			center = axis[0] + project(self.position[0]-axis[0], axis[1])
			cyl = generation.extrusion(
						axis[1]*size, 
						web(primitives.Circle(
								(center-axis[1]*size*0.5, axis[1]), 
								size/4, 
								resolution=('div', 16),
						)))
			l = len(cyl.points)
			v = junc - center
			v = normalize(noproject(v,axis[1]))
			if glm.any(isnan(v)):
				v,_,_ = dirbase(axis[1])
			p = center + v*size/4
			cyl.points.append(p)
			cyl.points.append(p + axis[1]*size*cornersize)
			cyl.points.append(p + v*size*cornersize)
			cyl.points.append(junc)
			return Scheme(
					cyl.points, 
					cyl.faces, 
					[(l,l+1,l+2)], 
					(	[(l, l+2), (l+2, l+3)] 
					+	[(i-1,i) for i in range(1,l//2)] 
					+	[(i-1,i) for i in range(l//2+1,l)] 
					+	[(l//2-1,0), (l-1,l//2)]),
					)
		elif solid is self.solids[1]:
			radius = size/4
			axis = self.axis[1]
			center = axis[0] + project(self.position[1]-axis[0], axis[1])
			attach = side = axis[1] * size
			return Scheme([center-side, center+side, center+attach, junc], [], [], [(0,1), (2,3)])


class Ball(Joint):
	''' Joint for rotation all around a point.
	
		classical definition: Ball (point)
		the initial state doen't require more data
		the class holds a point for each side
	'''
	def __init__(self, s1, s2, p1, p2=None):
		self.solids = (s1,s2)
		self.points = (p1, p2 or p1)
		self.position = self.points
	
	slvvars = 'solids',
	def fit(self):
		p0 = self.solids[0].orientation*self.points[0] + self.solids[0].position
		p1 = self.solids[1].orientation*self.points[1] + self.solids[1].position
		return distance(p0, p1)**2
	
	def corrections(self):
		p0 = self.solids[0].orientation*self.points[0] + self.solids[0].position
		p1 = self.solids[1].orientation*self.points[1] + self.solids[1].position
		return (
			Screw(p1-p0, vec3(0), p0),
			Screw(p0-p1, vec3(0), p1),
			) # force torsor
	
	def scheme(self, solid, size, junc=vec3(0,0,1)):
		if solid is self.solids[0]:
			r = 0.3*size
			p = self.points[0]
			sph = generation.icosphere(p, r, resolution=('rad',0.2))
			z = normalize(junc-p)
			if not isfinite(z):
				z = vec3(0,0,1)
			l = len(sph.points)
			sph.points.append(p + r*z)
			sph.points.append(p + r*z + 0.1*size*z)
			sph.points.append(junc)
			return Scheme(sph.points, sph.faces, [], [(l+0, l+1), (l+1, l+2)])
		elif solid is self.solids[1]:
			r = 0.4*size
			angle = 1.2
			p = self.points[1]
			z = normalize(junc-p)
			if not isfinite(z):		z = vec3(0,0,1)
			x,y,z = dirbase(z)
			prof = primitives.ArcCentered((p,y), p+z*r, p+r*(x*sin(angle)+z*cos(angle)))
			sph = generation.revolution(2*pi, (p,z), prof)
			sph.finish()
			l = len(sph.points)
			sph.points.append(p + r*z)
			sph.points.append(p + r*z + 0.1*size*z)
			sph.points.append(junc)
			return Scheme(sph.points, sph.faces, [], 
					[(l+0, l+1), (l+1, l+2)] + list(sph.outlines_unoriented()) )
from madcad.nprint import nprint
class Punctiform(Joint):
	''' Joint for rotation all around a point belonging to a plane.
	
		classical definition: Punctiform/Sphere-Plane (axis)
		the initial state does't require more data
		the class holds a normal axis for the plane side, and a point for the sphere side (that defaults to the axis'origin)
	'''
	def __init__(self, s1, s2, a1, p2=None):
		self.solids = (s1,s2)
		self.axis = a1
		self.point = p2 or a1[0]
		self.position = (self.axis[0], self.point)
	
	slvvars = 'solids',
	def fit(self):
		a = solidtransform_axis(self.solids[0], self.axis)
		p = self.solids[1].orientation*self.point + self.solids[1].position
		return project(p-a[0], a[1]) **2
	
	def corrections(self):
		a = solidtransform_axis(self.solids[0], self.axis)
		p = self.solids[1].orientation*self.point + self.solids[1].position
		t = Screw(project(p-a[0], a[1]), vec3(0), p)
		return (t, -t) # force torsor
	
	def scheme(self, solid, size, junc):
		r = 0.2*size
		if solid is self.solids[0]:
			center, normal = self.axis
			if dot(junc-center, normal) < 0:	normal = -normal
			x,y,z = dirbase(normal)
			c = center + r*z
			x *= size*0.7
			y *= size*0.7
			return Scheme(
				[c+(x+y), c+(-x+y), c+(-x-y), c+(x-y), c, c+cornersize*x, c+cornersize*z, junc],
				[(0,1,2), (0,2,3)],
				[(4,5,6)],
				[(0,1),(1,2),(2,3),(3,0),(4,6),(6,7)],
				)
		elif solid is self.solids[1]:
			sph = generation.icosphere(self.point, r, resolution=('rad', 0.2))
			z = normalize(junc-self.point)
			l = len(sph.points)
			sph.points.append(self.point + z*r)
			sph.points.append(junc)
			return Scheme(sph.points, sph.faces, [], [(l+0, l+1)])


class Gear(Joint):
	''' Gear interaction between two solids.
	
		`ratio` is the factor from the rotation of s1 to the rotation of s2
		The two pinions are considered circular, but no assumption is made on their radius or relative inclination.
		The interaction is symetric.
	'''
	def __init__(self, s1, s2, ratio, a1, a2=None, position=None):
		self.ratio = ratio
		self.solids = (s1, s2)
		self.axis = (a1, a2 or a1)
		self.position = position or (self.axis[0][0], self.axis[1][0])
	
	slvvars = 'solids',
	def fit(self):
		#t0 = dot(angleAxis(self.solids[0].orientation)*axis(self.solids[0].orientation), self.axis[0][1])
		#t1 = dot(angleAxis(self.solids[1].orientation)*axis(self.solids[1].orientation), self.axis[1][1])
		#return (t0 - t1) **2
		indev
	
	def corrections(self):
		b0 = solidtransform_base(self.solids[0], (self.axis[0][0], *dirbase(self.axis[0][1])))
		b1 = solidtransform_base(self.solids[1], (self.axis[1][0], *dirbase(self.axis[1][1])))
		# direction of the contact point from the center
		v = b1[0]-b0[0]
		# get the rotation of both solids relatively to this orientation
		t0, t1 = atan2(dot(v,b0[2]), dot(v,b0[1])), atan2(dot(v,b1[2]), dot(v,b1[1]))
		# correct cyclic orientation reset of the fastest solid
		if abs(self.ratio) < 1:
			m = (t1/self.ratio - t0 +pi) %(2*pi) -pi
		else:
			p = (t1 - t0*self.ratio +pi) %(2*pi) -pi
			m = p/self.ratio
		# m is the rotation delta for solid 0
		return (
			Screw(vec3(0), -m * b0[3], b0[0]),
			Screw(vec3(0), m*self.ratio * b1[3], b1[0]),
			)
		
	def scheme(self, solid, size, junc):
		if solid is self.solids[0]:		i0,i1,n = 0,1,self.ratio
		elif solid is self.solids[1]:	i0,i1,n = 1,0,1/self.ratio
		else:	return
		
		a0, a1 = solidtransform_axis(self.solids[0],self.axis[0]), solidtransform_axis(self.solids[1],self.axis[1])
		x,y,z = dirbase(self.axis[i0][1], align=junc-self.axis[i0][0])
		o = self.axis[i0][0] + project(self.position[i0]-self.axis[i0][0], z)
		
		# radial vector
		rl = length(noproject(a0[0]-a1[0], a1[1])) / (dot(a0[1],a1[1]) - 1/self.ratio)
		if solid is self.solids[1]:
			rl /= abs(self.ratio)
		r = o + x * rl
		
		# tooth depth vector
		angle = min(1, dot(	a0[1] if solid is self.solids[0] else a1[1], 
						normalize(mix(a0[1], a1[1], abs(self.ratio)/(1+abs(self.ratio))))
						))
		d = angle*z + sqrt(1-angle**2)*x
		b = size*0.4 * d
		w = size*0.08 * cross(d,y)
		if self.ratio > 0 and (	abs(self.ratio) > 1 and solid is self.solids[0] 
							or	abs(self.ratio) < 1 and solid is self.solids[1]):
			b, w = -b, -w
		
		profile = Web([r+b+w, r+b, r+b, r, r-b, r-b, r-b+w], [(0,1),(2,3),(3,4),(5,6)])
		surf = generation.revolution(2*pi, self.axis[i0], profile, resolution=('rad',0.1))
		l = len(profile.points)
		sch = Scheme(surf.points, surf.faces, [], [(i,i+l) for i in range(3, len(surf.points)-l, l)])
		sch.extend(Scheme([junc, mix(o,r,0.8), r], [], [], [(0,1),(1,2)]))
		return sch


class Helicoid(Joint):
	''' Screw a solid into an other.
	
		`step` is the translation distance needed for that one solid turn by 2*pi around the other
		The interaction is symetric.
	'''
	def __init__(self, s0, s1, step, b0, b1=None, position=None):
		self.step = step	# m/tr
		self.solids = s0, s1
		if isaxis(b0):			b0 = b0[0], *dirbase(b0[1])[:2]
		if b1 and isaxis(b1):	b1 = b1[0], *dirbase(b1[1])[:2]
		self.bases = b0, b1 or b0
		self.position = position or (self.bases[0][0], self.bases[1][0])
		
	slvvars = 'solids',
	def fit(self):
		indev

	def corrections(self):
		(o0,x0,y0), (o1,x1,y1) = solidtransform_base(self.solids[0], self.bases[0]), solidtransform_base(self.solids[1], self.bases[1])
		z0, z1 = cross(x0,y0), cross(x1,y1)
		z = normalize(z0 + z1)
		delta = o1 - o0
		pos = dot(-delta,z)
		angle = atan2(dot(y1,x0), dot(x1,x0))
		
		gap_angle = (pos%self.step)/self.step * 2*pi - angle
		if gap_angle > pi:	gap_angle -= 2*pi
		if abs(gap_angle) > 0.5:	gap_angle = 0
		gap_pos = (angle/(2*pi)*self.step - pos) % self.step
		if gap_pos > self.step/2:	gap_pos -= self.step
		
		r = cross(z0, z1) + gap_angle*z
		t = gap_pos*z
		return Screw(t+noproject(delta,z0), r, o0), Screw(-t-noproject(delta,z1), -r, o1)	# force torsor

	def scheme(self, solid, size, junc):
		if solid is self.solids[0]:
			radius = size/4
			o,x,y = self.bases[0]
			z = cross(x,y)
			center = o + project(self.position[0]-o, z)
			cyl = generation.extrusion(
						z*size, 
						web(primitives.Circle(
								(center-z*size*0.5, z), 
								radius, 
								resolution=('div', 16),
						)))
			l = len(cyl.points)
			v = junc - center
			v = normalize(noproject(v,z))
			if glm.any(isnan(v)):
				v,_,_ = dirbase(z)
			p = center + v*size/4
			cyl.points.append(p)
			cyl.points.append(p + z*size*cornersize)
			cyl.points.append(p + v*size*cornersize)
			cyl.points.append(junc)
			
			return Scheme(
					cyl.points, 
					cyl.faces, 
					[(l,l+1,l+2)], 
					(	[(l, l+2), (l+2, l+3)] 
					+	[(i-1,i) for i in range(1,l//2)] 
					+	[(i-1,i) for i in range(l//2+1,l)] 
					+	[(l//2-1,0), (l-1,l//2)]),
					)
					
		elif solid is self.solids[1]:
			radius = size/6
			o,x,y = self.bases[1]
			z = cross(x,y)
			center = o + project(self.position[1]-o, z)
			attach = side = z * size
			
			heli = []
			div = 32
			a = 4*pi
			for i in range(-div,div+1):
				t = i/div
				heli.append( radius*(cos(a*t)*x + sin(a*t)*y) + center + side*t )
			heli.extend([center+side, junc])
			heli = web(Wire(heli))
			return Scheme(heli.points, [], [], heli.edges)

